__all__ = ('CommandBaseApplicationCommand',)

from scarletio import copy_docs

from .....discord.application_command import ApplicationCommand, ApplicationCommandTargetType
from .....discord.application_command.application_command_permission.constants import (
    APPLICATION_COMMAND_PERMISSION_OVERWRITE_MAX
)
from .....discord.user.user.helpers import _try_get_guild_id

from ...utils import SYNC_ID_GLOBAL, SYNC_ID_NON_GLOBAL, UNLOADING_BEHAVIOUR_DELETE, UNLOADING_BEHAVIOUR_INHERIT

from ..command_base import CommandBase


class CommandBaseApplicationCommand(CommandBase):
    """
    Base class for ``Slasher``'s application commands.
    
    Attributes
    ----------
    _exception_handlers : `None`, `list` of `CoroutineFunction`
        Exception handlers added with ``.error`` to the interaction handler.
        
        Same as ``Slasher._exception_handlers``.
    
    _parent_reference : `None`, ``WeakReferer`` to ``Slasher``
        Reference to the slasher application command's parent.
    
    name : `str`
        Application command name. It's length can be in range [1:32].
    
    _permission_overwrites : `None`, `dict` of (`int`, `list` of ``ApplicationCommandPermissionOverwrite``)
        Permission overwrites applied to the slash command.

    _registered_application_command_ids : `None`, `dict` of (`int`, `int`) items
        The registered application command ids, which are matched by the command's schema.
        
        If empty set as `None`, if not then the keys are the respective guild's id and the values are the application
        command id.
    
    _schema : `None`, ``ApplicationCommand``
        Internal slot used by the ``.get_schema`` method.
    
    _unloading_behaviour : `int`
        Behaviour what describes what should happen when the command is removed from the slasher.
        
        Can be any of the following:
        
        +-------------------------------+-------+
        | Respective name               | Value |
        +-------------------------------+-------+
        | UNLOADING_BEHAVIOUR_DELETE    | 0     |
        +-------------------------------+-------+
        | UNLOADING_BEHAVIOUR_KEEP      | 1     |
        +-------------------------------+-------+
        | UNLOADING_BEHAVIOUR_INHERIT   | 2     |
        +-------------------------------+-------+
    
    global_ : `bool`
        Whether the command is a global command.
        
        Global commands have their``.guild_ids`` set as `None`.
    
    guild_ids : `None`, `set` of `int`
        The ``Guild``'s id to which the command is bound to.
    
    nsfw : `None`, `bool`
        Whether the application command is only allowed in nsfw channels.
    
    required_permissions : `None`, ``Permission``
        The required permissions to use the application command inside of a guild.
    
    Class Attributes
    ----------------
    COMMAND_COMMAND_NAME : `str`
        The command's name defining parameter's name.
    
    COMMAND_PARAMETER_NAMES : `tuple` of `str`
        All parameters names accepted by ``.__new__``
    
    COMMAND_NAME_NAME : `str`
        The command's "command" defining parameter's name.
    """
    __slots__ = (
        '_permission_overwrites', '_registered_application_command_ids', '_schema', '_unloading_behaviour',
        'allow_in_dm', 'default', 'global_', 'guild_ids', 'nsfw', 'required_permissions'
    )
    
    COMMAND_PARAMETER_NAMES = (
        *CommandBase.COMMAND_PARAMETER_NAMES,
        'delete_on_unload',
        'guild',
        'is_global',
        'nsfw',
        'required_permissions',
    )
    
    @property
    def target(self):
        """
        Returns command's target type.
        
        This property is a placeholder for subclasses which actually implement it.
        
        Returns
        -------
        target ``ApplicationCommandTargetType``
        """
        return ApplicationCommandTargetType.none
    
    
    @target.setter
    def target(self, value):
        pass
    
    
    @property
    def description(self):
        """
        Returns the command's description.
        
        This property is a placeholder for subclasses which actually implement it.
        
        Returns
        -------
        description `None`, `str`
        """
        return None
    
    
    @description.setter
    def description(self, value):
        return
    
    
    @copy_docs(CommandBase.copy)
    def copy(self):
        new = CommandBase.copy(self)
        
        # _permission_overwrites
        permission_overwrites = self._permission_overwrites
        if (permission_overwrites is not None):
            permission_overwrites = {
                guild_id: permission_overwrite.copy() for
                guild_id, permission_overwrite in permission_overwrites.items()
            }
        new._permission_overwrites = permission_overwrites
        
        # _registered_application_command_ids
        new._registered_application_command_ids = None
        
        # _schema
        new._schema = None
        
        # _unloading_behaviour
        new._unloading_behaviour = self._unloading_behaviour
        
        # allow_in_dm
        new.allow_in_dm = self.allow_in_dm
        
        # guild_ids
        guild_ids = self.guild_ids
        if (guild_ids is not None):
            guild_ids = guild_ids.copy()
        new.guild_ids = guild_ids
        
        # global_
        new.global_ = self.global_
        
        # nsfw
        new.nsfw = self.nsfw
        
        # required_permissions
        new.required_permissions = self.required_permissions
        
        return new
    
    
    @copy_docs(CommandBase.__hash__)
    def __hash__(self):
        hash_value = CommandBase.__hash__(self)
        
        # _permission_overwrites
        permission_overwrites = self._permission_overwrites
        if (permission_overwrites is not None):
            
            hash_value ^= len(permission_overwrites) << 8
            
            for permission_overwrite in permission_overwrites:
                hash_value ^= hash(permission_overwrite)
        
        # _registered_application_command_ids
        # Internal Field
        
        # _unloading_behaviour
        hash_value ^= (self._unloading_behaviour + 1) << 12
        
        # guild_ids
        guild_ids = self.guild_ids
        if (guild_ids is not None):
            hash_value ^= len(guild_ids) << 18
            
            for guild_id in guild_ids:
                hash_value ^= guild_id
        
        # nsfw
        nsfw = self.nsfw
        if (nsfw is not None):
            hash_value ^ nsfw << 22
        
        # required_permissions
        required_permissions = self.required_permissions
        if (required_permissions is not None):
            hash_value ^= required_permissions
        
        return hash_value
    
    
    @copy_docs(CommandBase._is_equal_same_type)
    def _is_equal_same_type(self, other):
        if not CommandBase._is_equal_same_type(self, other):
            return False
        
        # _permission_overwrites
        if self._permission_overwrites != other._permission_overwrites:
            return False
        
        # _registered_application_command_ids
        # Internal field

        # _unloading_behaviour
        if self._unloading_behaviour != other._unloading_behaviour:
            return False
        
        # allow_in_dm
        if self.allow_in_dm != other.allow_in_dm:
            return False
        
        # guild_ids
        if self.guild_ids != other.guild_ids:
            return False
        
        # global
        if self.global_ != other.global_:
            return False
        
        # nsfw
        if self.nsfw != other.nsfw:
            return False
        
        # required_permissions
        if self.required_permissions != other.required_permissions:
            return False
        
        return True
    
    
    @copy_docs(CommandBase._cursed_repr_builder)
    def _cursed_repr_builder(self):
        for repr_parts in CommandBase._cursed_repr_builder(self):
            
            repr_parts.append(', type = ')
            guild_ids = self.guild_ids
            if guild_ids is None:
                if self.global_:
                    type_name = 'global'
                else:
                    type_name = 'non-global'
            else:
                type_name = 'guild bound'
            
            repr_parts.append(type_name)
            
            yield repr_parts
            
            allow_in_dm = self.allow_in_dm
            if (allow_in_dm is not None):
                repr_parts.append(', allow_in_dm = ')
                repr_parts.append(repr(allow_in_dm))
            
            nsfw = self.nsfw
            if (nsfw is not None):
                repr_parts.append(', nsfw = ')
                repr_parts.append(repr(nsfw))
            
            required_permissions = self.required_permissions
            if (required_permissions is not None):
                repr_parts.append(', required_permissions = ')
                repr_parts.append(repr(required_permissions))
            
            unloading_behaviour = self._unloading_behaviour
            if unloading_behaviour != UNLOADING_BEHAVIOUR_INHERIT:
                repr_parts.append(', unloading_behaviour = ')
                if unloading_behaviour == UNLOADING_BEHAVIOUR_DELETE:
                    unloading_behaviour_name = 'delete'
                else:
                    unloading_behaviour_name = 'keep'
                
                repr_parts.append(unloading_behaviour_name)
            
            if (guild_ids is not None):
                repr_parts.append(', guild_ids = ')
                repr_parts.append(repr(guild_ids))
    
    
    async def invoke_auto_completion(self, client, interaction_event, auto_complete_option):
        """
        Calls the auto completion function of the slasher application command.
        
        This method is a coroutine.
        
        Parameters
        ----------
        client : ``Client``
            The respective client who received the event.
        interaction_event : ``InteractionEvent``
            The received interaction event.
        auto_complete_option : ``ApplicationCommandAutocompleteInteraction``
            The option to autocomplete.
        """
        pass
    
    # ---- Mention ----
    
    @property
    @copy_docs(CommandBase.mention)
    def mention(self):
        return self._build_mention(self._get_application_id_for_mention(), None)
    
    
    @copy_docs(CommandBase.mention_at)
    def mention_at(self, guild):
        return self._build_mention(self._get_application_id_for_mention_at(guild), None)
    
    
    def _mention_recursive(self, *sub_command_names):
        """
        Returns the application command's mention.
        
        Parameters
        ----------
        *sub_command_names : `str`
            Already included sub-command name stack to mention.
        
        Returns
        -------
        mention : `str`
        """
        return self._build_mention(self._get_application_id_for_mention(), sub_command_names)
    
    
    def _mention_at_recursive(self, guild, *sub_command_names):
        """
        Returns the application command's mention.
        
        Parameters
        ----------
        guild : ``Guild``, `int`
            The guild to mention the command at.
        
        *sub_command_names : `str`
            Already included sub-command name stack to mention.
        
        Returns
        -------
        mention : `str`
        """
        return self._build_mention(self._get_application_id_for_mention_at(guild), sub_command_names)
    
    
    def _get_application_id_for_mention(self):
        """
        Gets the application command's identifier to mention.
        
        This function is used by ``.mention`` and works if the application command is global or if it is added only
        to one guild.
        
        Returns
        -------
        application_command_id : `int`
        """
        registered_application_command_ids = self._registered_application_command_ids
        if (registered_application_command_ids is None) or (len(registered_application_command_ids) > 1):
            application_command_id = 0
        else:
            try:
                application_command_id = registered_application_command_ids[SYNC_ID_GLOBAL]
            except KeyError:
                application_command_id = next(iter(registered_application_command_ids.values()))
        
        return application_command_id
    
    
    def _get_application_id_for_mention_at(self, guild):
        """
        Gets the application command's identifier to mention at the specified guild.
        
        This function is used by ``.mention_at``.
        
        Parameters
        ----------
        guild : ``Guild``, `int`
            The guild to mention the application command at.
        
        Returns
        -------
        application_command_id : `int`
        """
        registered_application_command_ids = self._registered_application_command_ids
        if registered_application_command_ids is None:
            application_command_id = 0
        else:
            if self.global_:
                application_command_id = registered_application_command_ids.get(SYNC_ID_GLOBAL, 0)
            else:
                application_command_id = registered_application_command_ids.get(_try_get_guild_id(guild), 0)
        
        return application_command_id
    
    
    def _build_mention(self, application_command_id, sub_command_name_stack):
        """
        Builds the application command's mention.
        
        Parameters
        ----------
        application_command_id : `int`
            The respective application command's identifier.
        
        sub_command_name_stack : `None`, `tuple` of `str`
            Additional sub command names to include in the mention.
        
        Returns
        -------
        mention : `str`
        """
        mention_parts = []
        
        if application_command_id:
            mention_parts.append('<')
            
        mention_parts.append('/')
        mention_parts.append(self.name)
        
        if (sub_command_name_stack is not None):
            for sub_command_name in sub_command_name_stack:
                mention_parts.append(' ')
                mention_parts.append(sub_command_name)
        
        if application_command_id:
            mention_parts.append(':')
            mention_parts.append(str(application_command_id))
            mention_parts.append('>')
        
        return ''.join(mention_parts)
    
    # ---- Utility ----
    
    def get_real_command_count(self):
        """
        Gets the real command count of the command. This includes every sub attached to it as well.
        
        Returns
        -------
        real_command_count: `int`
        """
        return 1
    
    @property
    def interactions(self):
        """
        Enables you to add sub-commands or sub-categories to the command.
        
        Raises
        ------
        RuntimeError
            The ``CommandBaseApplicationCommand`` is not a category.
        """
        raise RuntimeError(
            f'The {self.__class__.__name__} is not a category.'
        )
    
    
    def autocomplete(self, parameter_name, *parameter_names, function = None):
        """
        Registers an auto completer function to the application command.
        
        Can be used as a decorator, as:
        
        ```py
        @bot.interactions(is_global = True)
        async def buy(
            item: ('str', 'Select an item to buy.'),
        ):
            return 'Great success.'
        
        AUTO_COMPLETE_CHOICES = (
            'cake',
            'shrimp fry',
        )
        
        @buy.autocomplete('item')
        async def autocomplete_item_parameter(value):
            if value is None:
                return AUTO_COMPLETE_CHOICES[:20]
            
            value = value.lower()
            
            return [choice for choice in AUTO_COMPLETE_CHOICES if choice.startswith(value)]
        ```
        
        Parameters
        ----------
        parameter_name : `str`
            The parameter's name.
        *parameter_names : `str`
            Additional parameter names to autocomplete
        function : `None`, `async-callable` = `None`, Optional (Keyword only)
            The function to register as auto completer.
        
        Returns
        -------
        function / wrapper : `async-callable`, `functools.partial`
            The registered function if given or a wrapper to register the function with.
        
        Raises
        ------
        RuntimeError
            - If the parameter already has a auto completer defined.
            - If the application command function has no parameter named, like `parameter_name`.
            - If the parameter cannot be auto completed.
        TypeError
            If `function` is not an asynchronous.
        """
        raise RuntimeError(
            f'{self.__class__.__name__} is not auto-completable.'
        )
    
    
    # ---- Permission overwrites ----

    def add_permission_overwrite(self, guild_id, permission_overwrite):
        """
        Adds an overwrite to the slash command.
        
        Parameters
        ----------
        guild_id : `int`
            The guild's id where the overwrite will be applied.
        permission_overwrite : ``ApplicationCommandPermissionOverwrite``, `None`
            The permission overwrite to add
        
        Raises
        ------
        AssertionError
            - Each command in each guild can have up to `10` overwrite, which is already reached.
        """
        permission_overwrites = self._permission_overwrites
        if permission_overwrites is None:
            self._permission_overwrites = permission_overwrites = {}
        
        permission_overwrites_for_guild = permission_overwrites.get(guild_id, None)
        
        if __debug__:
            if (
                (permission_overwrites_for_guild is not None) and
                (len(permission_overwrites_for_guild) >= APPLICATION_COMMAND_PERMISSION_OVERWRITE_MAX)
            ):
                raise AssertionError(
                    f'`Each command in each guild can have up to '
                    f'{APPLICATION_COMMAND_PERMISSION_OVERWRITE_MAX} permission overwrites which is already reached.'
                )
        
        if (permission_overwrites_for_guild is not None) and (permission_overwrite is not None):
            target_id = permission_overwrite.target_id
            for index in range(len(permission_overwrites_for_guild)):
                iter_permission_overwrites = permission_overwrites_for_guild[index]
                
                if iter_permission_overwrites.target_id != target_id:
                    continue
                
                if permission_overwrite.allow == iter_permission_overwrites.allow:
                    return
                
                del permission_overwrites_for_guild[index]
                
                if permission_overwrites_for_guild:
                    return
                
                permission_overwrites[guild_id] = None
                return
        
        if permission_overwrite is None:
            if permission_overwrites_for_guild is None:
                permission_overwrites[guild_id] = None
        else:
            if permission_overwrites_for_guild is None:
                permission_overwrites[guild_id] = permission_overwrites_for_guild = []
            
            permission_overwrites_for_guild.append(permission_overwrite)
    
    
    def get_permission_overwrites_for(self, guild_id):
        """
        Returns the slash command's permissions overwrites for the given guild.
        
        Returns
        -------
        permission_overwrites : `None`, `list` of ``ApplicationCommandPermissionOverwrite``
            Returns `None` instead of an empty list.
        """
        permission_overwrites = self._permission_overwrites
        if (permission_overwrites is not None):
            return permission_overwrites.get(guild_id, None)
    
    # ---- Sync ----
    
    def _get_permission_sync_ids(self):
        """
        Gets the permission overwrite guild id-s which should be synced.
        
        Returns
        -------
        permission_sync_ids : `set` of `int`
        """
        permission_sync_ids = set()
        guild_ids = self.guild_ids
        # If the command is guild bound, sync it in every guild, if not, then sync it in every guild where it has an
        # a permission overwrite.
        if (guild_ids is None):
            permission_overwrites = self._permission_overwrites
            if (permission_overwrites is not None):
                permission_sync_ids.update(permission_overwrites.keys())
        else:
            permission_sync_ids.update(guild_ids)
        
        return permission_sync_ids


    def _register_guild_and_application_command_id(self, guild_id, application_command_id):
        """
        Registers an application command's identifier to the ``SlashCommand`.
        
        Parameters
        ----------
        application_command_id : `int`
            The application command's identifier.
        guild_id : `int`
            The guild where the application command is in.
        """
        registered_application_command_ids = self._registered_application_command_ids
        if registered_application_command_ids is None:
            registered_application_command_ids = self._registered_application_command_ids = {}
        
        registered_application_command_ids[guild_id] = application_command_id
    
    
    def _unregister_guild_and_application_command_id(self, guild_id, application_command_id):
        """
        Unregisters an application command's identifier from the ``SlashCommand`.
        
        Parameters
        ----------
        guild_id : `int`
            The guild's id, where the application command is in.
        application_command_id : `int`
            The application command's identifier.
        """
        registered_application_command_ids = self._registered_application_command_ids
        if registered_application_command_ids is not None:
            try:
                maybe_application_command_id = registered_application_command_ids[guild_id]
            except KeyError:
                pass
            else:
                if maybe_application_command_id == application_command_id:
                    del registered_application_command_ids[guild_id]
                    
                    if not registered_application_command_ids:
                        self._registered_application_command_ids = None
    
    
    def _pop_command_id_for(self, guild_id):
        """
        Pops the given application command id from the command for the respective guild.
        
        Parameters
        ----------
        guild_id : `int`
            A guild's identifier.
        
        Returns
        -------
        application_command_id : `int`
            The popped application command's identifier. Returns `0` if nothing is matched.
        """
        registered_application_command_ids = self._registered_application_command_ids
        if registered_application_command_ids is None:
            application_command_id = 0
        else:
            application_command_id = registered_application_command_ids.pop(guild_id, 0)
        
        return application_command_id
    
    
    def _iter_application_command_ids(self):
        """
        Iterates over all the registered application command id-s added to the slash command.
        
        This method is a generator, what should be used inside of a `for` loop.
        
        Yields
        ------
        application_command_id : `int`
        """
        registered_application_command_ids = self._registered_application_command_ids
        if (registered_application_command_ids is not None):
            yield from registered_application_command_ids.values()
    
    
    def _exhaust_application_command_ids(self):
        """
        Iterates over all the registered application command id-s added to the slash command and removes them.
        
        This method is a generator, what should be used inside of a `for` loop.
        
        Yields
        ------
        application_command_id : `int`
        """
        registered_application_command_ids = self._registered_application_command_ids
        if (registered_application_command_ids is not None):
            while registered_application_command_ids:
                guild_id, application_command_id = registered_application_command_ids.popitem()
                yield application_command_id
            
            self._registered_application_command_ids = None
    
    
    def _iter_sync_ids(self):
        """
        Iterates over all the respective sync ids of the command. If the command is a guild bound command, then will
        iterate over it's guild's id-s.
        
        This method is a generator, what should be used inside of a `for` loop.
        
        Yields
        ------
        sync_id : `int`
        """
        if self.global_:
            yield SYNC_ID_GLOBAL
            return
        
        guild_ids = self.guild_ids
        if guild_ids is None:
            yield SYNC_ID_NON_GLOBAL
            return
        
        yield from guild_ids
    
    
    def _iter_guild_ids(self):
        """
        Iterates over all the guild identifiers used by the command.
        
        Yields
        ------
        guild_id : `int`
        """
        registered_application_command_ids = self._registered_application_command_ids
        if registered_application_command_ids is not None:
            for sync_id in registered_application_command_ids:
                if sync_id > (1 << 22):
                    yield sync_id
    
    # ---- Schema ----

    def get_schema(self):
        """
        Returns an application command schema representing the slash command.
        
        Returns
        -------
        schema : ``ApplicationCommand``
        """
        schema = self._schema
        if schema is None:
            schema = self._schema = self.as_schema()
        
        return schema
    
    
    def as_schema(self):
        """
        Creates a new application command schema representing the slash command.
        
        Returns
        -------
        schema : ``ApplicationCommand``
        """
        schema = ApplicationCommand(
            self.name,
            self.description,
            allow_in_dm = self.allow_in_dm,
            options = self._get_schema_options(),
            nsfw = self.nsfw,
            required_permissions = self.required_permissions,
            target_type = self.target,
        )
        
        parent_reference = self._parent_reference
        if (parent_reference is not None):
            parent = parent_reference()
            if (parent is not None):
                schema = schema.with_translation(parent._translation_table)
        
        return schema
    
    
    def _get_schema_options(self):
        """
        Gets the schema options for the application command.
        
        Returns
        -------
        application_command_options : `None`, `list` of ``ApplicationCommandOption``
        """
        return None



__all__ = ('OrinUserBase',)

from scarletio import copy_docs

from ...bases import ICON_TYPE_ANIMATED_APNG, ICON_TYPE_NONE, IconSlot
from ...http import urls as module_urls

from .fields import (
    parse_banner_color, parse_discriminator, parse_display_name, parse_flags, validate_banner_color,
    validate_discriminator, validate_display_name, validate_flags
)
from .flags import UserFlag
from .user_base import UserBase


class OrinUserBase(UserBase):
    """
    Base class for actual user entities, like oauth2 user and normal users.
    
    Attributes
    ----------
    avatar_hash : `int`
        The user's avatar's hash in `uint128`.
    avatar_type : ``IconType``
        The user's avatar's type.
    avatar_decoration_hash : `int`
        The user's avatar decoration's hash in `uint128`.
    avatar_decoration_type : ``IconType``
        The user's avatar decoration's type.
    banner_color : `None`, ``Color``
        The user's banner color if has any.
    banner_hash : `int`
        The user's banner's hash in `uint128`.
    banner_type : ``IconType``
        The user's banner's type.
    discriminator : `int`
        The client's discriminator. Given to avoid overlapping names.
    display_name : `None`, `str`
        The user's non-unique display name.
    id : `int`
        The client's unique identifier number.
    flags : ``UserFlag``
        The user's flags.
    name : str
        The client's username.
    """
    __slots__ = ('banner_color', 'discriminator', 'display_name', 'flags')
    
    banner = IconSlot('banner', 'banner', module_urls.user_banner_url, module_urls.user_banner_url_as)
    avatar_decoration = IconSlot(
        'avatar_decoration',
        'avatar_decoration',
        module_urls.user_avatar_decoration_url,
        module_urls.user_avatar_decoration_url_as,
        animated_icon_type = ICON_TYPE_ANIMATED_APNG,
    )
    
    def __new__(
        cls,
        *,
        avatar = ...,
        avatar_decoration = ...,
        banner = ...,
        banner_color = ...,
        discriminator = ...,
        display_name = ...,
        flags = ...,
        name = ...,
    ):
        """
        Creates a new partial user with the given fields.
        
        Parameters
        ----------
        avatar : `None`, ``Icon``, `str`, `bytes-like`, Optional (Keyword only)
            The user's avatar.
        avatar_decoration : `None`, ``Icon``, `str`, `bytes-like`, Optional (Keyword only)
            The user's avatar decoration.
        banner : `None`, ``Icon``, `str`, `bytes-like`, Optional (Keyword only)
            The user's banner.
        banner_color : `None`, ``Color``, `int`, Optional (Keyword only)
            The user's banner color.
        discriminator : `str`, `int`, Optional (Keyword only)
            The user's discriminator.
        display_name : `None`, `str`, Optional (Keyword only)
            The user's non-unique display name.
        flags : `int`, ``UserFlag``, Optional (Keyword only)
            The user's flags.
        name : `str`, Optional (Keyword only)
            The user's name.
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        # avatar_decoration
        if avatar_decoration is ...:
            avatar_decoration = None
        else:
            avatar_decoration = cls.avatar_decoration.validate_icon(avatar_decoration, allow_data = True)
        
        # banner
        if banner is ...:
            banner = None
        else:
            banner = cls.banner.validate_icon(banner, allow_data = True)
        
        # banner_color
        if banner_color is ...:
            banner_color = None
        else:
            banner_color = validate_banner_color(banner_color)
        
        # discriminator
        if discriminator is ...:
            discriminator = 0
        else:
            discriminator = validate_discriminator(discriminator)
        
        # display_name
        if display_name is ...:
            display_name = None
        else:
            display_name = validate_display_name(display_name)
        
        # flags
        if flags is ...:
            flags = UserFlag()
        else:
            flags = validate_flags(flags)
        
        # Construct
        self = UserBase.__new__(
            cls,
            avatar = avatar,
            name = name,
        )
        self.avatar_decoration = avatar_decoration
        self.banner = banner
        self.banner_color = banner_color
        self.discriminator = discriminator
        self.display_name = display_name
        self.flags = flags
        return self
    
    
    @copy_docs(UserBase._update_attributes)
    def _update_attributes(self, data):
        UserBase._update_attributes(self, data)
        
        self._set_avatar_decoration(data)
        self._set_banner(data)
        self.banner_color = parse_banner_color(data)
        self.discriminator = parse_discriminator(data)
        self.display_name = parse_display_name(data)
        self.flags = parse_flags(data)
        
    
    @copy_docs(UserBase._difference_update_attributes)
    def _difference_update_attributes(self, data):
        old_attributes = UserBase._difference_update_attributes(self, data)
        
        # avatar_decoration
        self._update_avatar_decoration(data, old_attributes)
        
        # banner
        self._update_banner(data, old_attributes)
        
        # banner_color
        banner_color = parse_banner_color(data)
        if self.banner_color != banner_color:
            old_attributes['banner_color'] = self.banner_color
            self.banner_color = banner_color
        
        # discriminator
        discriminator = parse_discriminator(data)
        if self.discriminator != discriminator:
            old_attributes['discriminator'] = self.discriminator
            self.discriminator = discriminator
        
        # display_name
        display_name = parse_display_name(data)
        if self.display_name != display_name:
            old_attributes['display_name'] = self.display_name
            self.display_name = display_name
        
        # flags
        flags = parse_flags(data)
        if self.flags != flags:
            old_attributes['flags'] = self.flags
            self.flags = flags
        
        return old_attributes
    
    
    @copy_docs(UserBase._set_default_attributes)
    def _set_default_attributes(self):
        UserBase._set_default_attributes(self)
        self.avatar_decoration_hash = 0
        self.avatar_decoration_type = ICON_TYPE_NONE
        self.banner_color = None
        self.banner_hash = 0
        self.banner_type = ICON_TYPE_NONE
        self.discriminator = 0
        self.display_name = None
        self.flags = UserFlag()
    
    
    @copy_docs(UserBase.copy)
    def copy(self):
        new = UserBase.copy(self)
        new.avatar_decoration = self.avatar_decoration
        new.banner = self.banner
        new.banner_color = self.banner_color
        new.discriminator = self.discriminator
        new.display_name = self.display_name
        new.flags = self.flags
        return new
    
    
    def copy_with(
        self,
        *,
        avatar = ...,
        avatar_decoration = ...,
        banner = ...,
        banner_color = ...,
        discriminator = ...,
        display_name = ...,
        flags = ...,
        name = ...,
    ):
        """
        Copies the user with the given fields.
        
        Parameters
        ----------
        avatar : `None`, ``Icon``, `str`, `bytes-like`, Optional (Keyword only)
            The user's avatar.
        avatar_decoration : `None`, ``Icon``, `str`, `bytes-like`, Optional (Keyword only)
            The user's avatar decoration.
        banner : `None`, ``Icon``, `str`, `bytes-like`, Optional (Keyword only)
            The user's banner.
        banner_color : `None`, ``Color``, `int`, Optional (Keyword only)
            The user's banner color.
        discriminator : `str`, `int`, Optional (Keyword only)
            The user's discriminator.
        display_name : `None`, `str`, Optional (Keyword only)
            The user's non-unique display name.
        flags : `int`, ``UserFlag``, Optional (Keyword only)
            The user's flags.
        name : `str`, Optional (Keyword only)
            The user's name.
        
        Returns
        -------
        new : `instance<type<self>>`
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        # avatar_decoration
        if avatar_decoration is ...:
            avatar_decoration = self.avatar_decoration
        else:
            avatar_decoration = type(self).avatar_decoration.validate_icon(avatar_decoration, allow_data = True)
        
        # banner
        if banner is ...:
            banner = self.banner
        else:
            banner = type(self).banner.validate_icon(banner, allow_data = True)
        
        # banner_color
        if banner_color is ...:
            banner_color = self.banner_color
        else:
            banner_color = validate_banner_color(banner_color)
        
        # discriminator
        if discriminator is ...:
            discriminator = self.discriminator
        else:
            discriminator = validate_discriminator(discriminator)
        
        # display_name
        if display_name is ...:
            display_name = self.display_name
        else:
            display_name = validate_display_name(display_name)
        
        # flags
        if flags is ...:
            flags = self.flags
        else:
            flags = validate_flags(flags)
        
        # Construct
        new = UserBase.copy_with(
            self,
            avatar = avatar,
            name = name,
        )
        new.avatar_decoration = avatar_decoration
        new.banner = banner
        new.banner_color = banner_color
        new.discriminator = discriminator
        new.display_name = display_name
        new.flags = flags
        return new
    
    
    @copy_docs(UserBase._get_hash_partial)
    def _get_hash_partial(self):
        hash_value = UserBase._get_hash_partial(self)
        
        # avatar_decoration
        hash_value ^= hash(self.avatar_decoration)
        
        # banner
        hash_value ^= hash(self.banner)
        
        # banner_color
        banner_color = self.banner_color
        if (banner_color is not None):
            hash_value ^= 1 << 25
            hash_value ^= banner_color
        
        # discriminator
        hash_value ^= self.discriminator << 26
        
        # display_name
        display_name = self.display_name
        if (display_name is not None) and (display_name != self.name):
            hash_value ^= hash(display_name)
        
        # flags
        hash_value ^= self.flags
        
        return hash_value



__all__ = ()

from scarletio import include

from ..core import (
    AUTO_MODERATION_RULES, CHANNELS, EMOJIS, INTEGRATIONS, ROLES, SCHEDULED_EVENTS, STAGES, STICKERS, USERS
)

AuditLogEvent = include('AuditLogEvent')
Invite = include('Invite')

def target_converter_none(entry):
    return None


def target_converter_guild(entry):
    parent = entry.parent
    if (parent is not None):
        return parent.guild


def target_converter_channel(entry):
    target_id = entry.target_id
    if target_id:
        return CHANNELS.get(target_id, None)


def target_converter_user(entry):
    target_id = entry.target_id
    if target_id:
        return USERS.get(target_id, None)


def target_converter_role(entry):
    target_id = entry.target_id
    if target_id:
        return ROLES.get(target_id, None)


def target_converter_invite(entry):
    # every other data is at # change
    for change in entry.changes:
        if change.attribute_name != 'code':
            continue
        
        if entry.type is AuditLogEvent.invite_delete:
            code = change.before
        else:
            code = change.after
        break
    
    else:
        code = '' # malformed ?
    
    
    parent = entry.parent
    if (parent is None):
        guild = None
    else:
        guild = parent.guild
    
    return Invite.precreate(code, guild = guild)


def target_converter_webhook(entry):
    target_id = entry.target_id
    if target_id:
        return USERS.get(target_id, None)


def target_converter_emoji(entry):
    target_id = entry.target_id
    if target_id:
        return EMOJIS.get(target_id, None)


def target_converter_integration(entry):
    target_id = entry.target_id
    if target_id:
        return INTEGRATIONS.get(target_id, None)


def target_converter_stage(entry):
    target_id = entry.target_id
    if target_id:
        return STAGES.get(target_id, None)


def target_converter_sticker(entry):
    target_id = entry.target_id
    if target_id:
        return STICKERS.get(target_id, None)


def target_converter_scheduled_event(entry):
    target_id = entry.target_id
    if target_id:
        return SCHEDULED_EVENTS.get(target_id, None)


def target_converter_thread(entry):
    target_id = entry.target_id
    if target_id:
        return CHANNELS.get(target_id, None)


def target_converter_application_command(entry):
    target_id = entry.target_id
    if target_id:
        parent = entry.parent
        if (parent is not None):
            return parent.application_commands(target_id, None)


def target_converter_auto_moderation_rule(entry):
    target_id = entry.target_id
    if target_id:
        return AUTO_MODERATION_RULES.get(target_id, None)



__all__ = ('Component',)

from scarletio import RichAttributeErrorBaseType, copy_docs, export

from ..component_metadata import ComponentMetadataBase
from ..component_metadata.fields import (
    validate_button_style, validate_channel_types, validate_enabled, validate_label, validate_max_length,
    validate_max_values, validate_min_length, validate_min_values, validate_options, validate_placeholder,
    validate_required, validate_text_input_style, validate_url, validate_value
)
from ..shared_fields import validate_components, validate_custom_id, validate_emoji

from .fields import parse_type, put_type_into, validate_type
from .preinstanced import ComponentType


@export
class Component(RichAttributeErrorBaseType):
    """
    Represents a message component.
    
    Attributes
    ----------
    type : ``ComponentType``
        The component's type.
    metadata : ``ComponentMetadataBase``
        The component's metadata.
    """
    __slots__ = ('metadata', 'type',)
    
    def __new__(cls, component_type, **keyword_parameters):
        """
        Creates a component with the given parameters.
        
        Parameters
        ----------
        component_type : `int`, ``ComponentType``
            The component's type to create.
        
        **keyword_parameters : Keyword parameters
            Keyword parameters defining the component's fields.
        
        Other parameters
        ----------------
        button_style : ``ButtonStyle``, Optional (Keyword only)
            The component's style. Applicable for button components.
        
        channel_types : `None`, `tuple` of (``ChannelType``, `int`), Optional (Keyword only)
            The allowed channel types by the select.
        
        components : `None`, `tuple` of ``Component``, Optional (Keyword only)
            Sub-components.
        
        custom_id : `None`, `str`, Optional (Keyword only)
            Custom identifier to detect which component was clicked (or used) by the user.
        
        emoji : `None` ``Emoji``, Optional (Keyword only)
            Emoji of the component if applicable.
        
        enabled : `bool`, Optional (Keyword only)
            Whether the component is enabled.
        
        label : `None`, `str`, Optional (Keyword only)
            Label of the component.
        
        max_length : `int`, Optional (Keyword only)
            The maximal length of the inputted text.
        
        max_values : `int`, Optional (Keyword only)
            The maximal amount of options to select.
        
        min_length : `int`, Optional (Keyword only)
            The minimal length of the inputted text.
        
        min_values : `int`, Optional (Keyword only)
            The minimal amount of options to select.
        
        options : `None`, `tuple` of ``StringSelectOption``, Optional (Keyword only)
            Options of the component.
        
        placeholder : `None`, `str`, Optional (Keyword only)
            Placeholder of the select.
        
        required : `bool`, Optional (Keyword only)
            Whether the field is required to be fulfilled.
        
        text_input_style : ``TextInputStyle``, Optional (Keyword only)
            The text input's style.
        
        url : `None`, `str`, Optional (Keyword only)
            Url to redirect to.
        
        value : `None`, `str`, Optional (Keyword only)
            The input component's default value.
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
            - Extra or unused parameters.
        ValueError
            - If a parameter's value is incorrect.
        """
        component_type = validate_type(component_type)
        metadata = component_type.metadata_type.from_keyword_parameters(keyword_parameters)
        
        if keyword_parameters:
            raise TypeError(
                f'Extra or unused keyword parameters: {keyword_parameters!r}.'
            )
        
        self = object.__new__(cls)
        self.type = component_type
        self.metadata = metadata
        return self
    
    
    def __iter__(self):
        """
        Iterates over the sub components.
        
        This method is an iterable generator.
        
        Yields
        ------
        component
        """
        components = self.components
        if (components is not None):
            yield from components
    
    
    def __eq__(self, other):
        """Returns whether the two components are equal."""
        if type(self) is not type(other):
            return NotImplemented
        
        if self.type is not other.type:
            return False
        
        if self.metadata != other.metadata:
            return False
        
        return True
    
    
    def __repr__(self):
        """Returns the component's representation."""
        repr_parts = ['<', self.__class__.__name__]
        
        repr_parts.append(' type = ')
        
        component_type = self.type
        repr_parts.append(component_type.name)
        repr_parts.append(' ~ ')
        repr_parts.append(repr(component_type.value))
        
        repr_parts.append(', metadata = ')
        repr_parts.append(repr(self.metadata))
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    def __hash__(self):
        """Returns the component's hash value."""
        hash_value = 0
        
        # metadata
        hash_value ^= hash(self.metadata)
        
        # type
        hash_value ^= self.type.value
        
        return hash_value
    
    
    @classmethod
    def from_data(cls, data):
        """
        Creates a component from the given data.
        
        Parameters
        ----------
        data : `dict` of (`str`, `object`) items
            Component data.
        
        Returns
        -------
        self : `instance<cls>`
            Returns the created component.
        """
        component_type = parse_type(data)
        metadata = component_type.metadata_type.from_data(data)
        
        self = object.__new__(cls)
        self.metadata = metadata
        self.type = component_type
        return self
    
    
    def to_data(self, *, defaults = False):
        """
        Returns the component's json serializable representation.
        
        Parameters
        ----------
        defaults : `bool` = `False`, Optional (Keyword only)
            Whether fields with default values should be included as well.
        
        Returns
        -------
        data : `dict` of (`str`, `object`) items
        """
        # metadata
        data = self.metadata.to_data(defaults = defaults)
        
        # type
        put_type_into(self.type, data, defaults)
        
        return data
    
    
    def copy(self):
        """
        Copies the component.
        
        Returns
        -------
        new : `instance<type<self>>`
        """
        metadata = self.metadata.copy()
        
        new = object.__new__(type(self))
        new.metadata = metadata
        new.type = self.type
        return new
    
    
    def copy_with(self, *, component_type = ..., **keyword_parameters):
        """
        Copies the component with changing it's field.
        
        Parameters
        ----------
        component_type : `int`, ``ComponentType``, Optional (Keyword only)
            The component's type.
        
        **keyword_parameters : Keyword parameters
            Keyword parameters defining which fields should be changed.
        
        Other parameters
        ----------------
        button_style : ``ButtonStyle``, Optional (Keyword only)
            The component's style. Applicable for button components.
        
        channel_types : `None`, `tuple` of (``ChannelType``, `int`), Optional (Keyword only)
            The allowed channel types by the select.
        
        components : `None`, `tuple` of ``Component``, Optional (Keyword only)
            Sub-components.
        
        custom_id : `None`, `str`, Optional (Keyword only)
            Custom identifier to detect which component was clicked (or used) by the user.
        
        emoji : `None` ``Emoji``, Optional (Keyword only)
            Emoji of the component if applicable.
        
        enabled : `bool`, Optional (Keyword only)
            Whether the component is enabled.
        
        label : `None`, `str`, Optional (Keyword only)
            Label of the component.
        
        max_length : `int`, Optional (Keyword only)
            The maximal length of the inputted text.
        
        max_values : `int`, Optional (Keyword only)
            The maximal amount of options to select.
        
        min_length : `int`, Optional (Keyword only)
            The minimal length of the inputted text.
        
        min_values : `int`, Optional (Keyword only)
            The minimal amount of options to select.
        
        options : `None`, `tuple` of ``StringSelectOption``, Optional (Keyword only)
            Options of the component.
        
        placeholder : `None`, `str`, Optional (Keyword only)
            Placeholder of the select.
        
        required : `bool`, Optional (Keyword only)
            Whether the field is required to be fulfilled.
        
        text_input_style : ``TextInputStyle``, Optional (Keyword only)
            The text input's style.
        
        url : `None`, `str`, Optional (Keyword only)
            Url to redirect to.
        
        value : `None`, `str`, Optional (Keyword only)
            The input component's default value.
        
        Returns
        -------
        new : `instance<type<self>>`
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
            - Extra or unused parameters.
        ValueError
            - If a parameter's value is incorrect.
        """
        # component_type
        if component_type is ...:
            component_type = self.type
        else:
            component_type = validate_type(component_type)
        
        # metadata
        if component_type is self.type:
            metadata = self.metadata.copy_with_keyword_parameters(keyword_parameters)
        else:
            metadata = component_type.metadata_type.from_keyword_parameters(keyword_parameters)
        
        if keyword_parameters:
            raise TypeError(
                f'Extra or unused keyword parameters: {keyword_parameters!r}.'
            )
        
        # Construct
        new = object.__new__(type(self))
        new.metadata = metadata
        new.type = component_type
        return new
    
    
    # Field proxies
    
    @property
    @copy_docs(ComponentMetadataBase.button_style)
    def button_style(self):
        return self.metadata.button_style
    
    @button_style.setter
    def button_style(self, button_style):
        self.metadata.button_style = validate_button_style(button_style)
    
    
    @property
    @copy_docs(ComponentMetadataBase.channel_types)
    def channel_types(self):
        return self.metadata.channel_types
    
    @channel_types.setter
    def channel_types(self, channel_types):
        self.metadata.channel_types = validate_channel_types(channel_types)
    
    
    @property
    @copy_docs(ComponentMetadataBase.components)
    def components(self):
        return self.metadata.components
    
    @components.setter
    def components(self, components):
        self.metadata.components = validate_components(components)
    
    
    @property
    @copy_docs(ComponentMetadataBase.custom_id)
    def custom_id(self):
        return self.metadata.custom_id
    
    @custom_id.setter
    def custom_id(self, custom_id):
        self.metadata.custom_id = validate_custom_id(custom_id)
    
    
    @property
    @copy_docs(ComponentMetadataBase.emoji)
    def emoji(self):
        return self.metadata.emoji
    
    @emoji.setter
    def emoji(self, emoji):
        self.metadata.emoji = validate_emoji(emoji)
    
    
    @property
    @copy_docs(ComponentMetadataBase.enabled)
    def enabled(self):
        return self.metadata.enabled
    
    @enabled.setter
    def enabled(self, enabled):
        self.metadata.enabled = validate_enabled(enabled)
    
    
    @property
    @copy_docs(ComponentMetadataBase.label)
    def label(self):
        return self.metadata.label
    
    @label.setter
    def label(self, label):
        self.metadata.label = validate_label(label)
    
    
    @property
    @copy_docs(ComponentMetadataBase.max_length)
    def max_length(self):
        return self.metadata.max_length
    
    @max_length.setter
    def max_length(self, max_length):
        self.metadata.max_length = validate_max_length(max_length)
    
    
    @property
    @copy_docs(ComponentMetadataBase.max_values)
    def max_values(self):
        return self.metadata.max_values
    
    @max_values.setter
    def max_values(self, max_values):
        self.metadata.max_values = validate_max_values(max_values)
    
    
    @property
    @copy_docs(ComponentMetadataBase.min_length)
    def min_length(self):
        return self.metadata.min_length
    
    @min_length.setter
    def min_length(self, min_length):
        self.metadata.min_length = validate_min_length(min_length)
    
    
    @property
    @copy_docs(ComponentMetadataBase.min_values)
    def min_values(self):
        return self.metadata.min_values
    
    @min_values.setter
    def min_values(self, min_values):
        self.metadata.min_values = validate_min_values(min_values)
    
    
    @property
    @copy_docs(ComponentMetadataBase.options)
    def options(self):
        return self.metadata.options
    
    @options.setter
    def options(self, options):
        self.metadata.options = validate_options(options)
    
    
    @property
    @copy_docs(ComponentMetadataBase.placeholder)
    def placeholder(self):
        return self.metadata.placeholder
    
    @placeholder.setter
    def placeholder(self, placeholder):
        self.metadata.placeholder = validate_placeholder(placeholder)
    
    
    @property
    @copy_docs(ComponentMetadataBase.required)
    def required(self):
        return self.metadata.required
    
    @required.setter
    def required(self, required):
        self.metadata.required = validate_required(required)
    
    
    @property
    @copy_docs(ComponentMetadataBase.text_input_style)
    def text_input_style(self):
        return self.metadata.text_input_style
    
    @text_input_style.setter
    def text_input_style(self, text_input_style):
        self.metadata.text_input_style = validate_text_input_style(text_input_style)
    
    @property
    @copy_docs(ComponentMetadataBase.url)
    def url(self):
        return self.metadata.url
    
    @url.setter
    def url(self, url):
        self.metadata.url = validate_url(url)
    
    
    @property
    @copy_docs(ComponentMetadataBase.value)
    def value(self):
        return self.metadata.value
    
    @value.setter
    def value(self, value):
        self.metadata.value = validate_value(value)
    
    # Field proxy utilities
    
    @property
    def style(self):
        """
        Returns the component's style.
        
        Returns
        -------
        style : `None`, ``ButtonStyle``, ``TextInputStyle``
        """
        component_type = self.type
        if component_type is ComponentType.button:
            return self.button_style
        
        if component_type is ComponentType.text_input:
            return self.text_input_style
    
    @style.setter
    def style(self, style):
        component_type = self.type
        if component_type is ComponentType.button:
            self.metadata.button_style = validate_button_style(style)
            return
        
        if component_type is ComponentType.text_input:
            self.metadata.text_input_style = validate_text_input_style(style)
            return
        
        raise TypeError(
            f'`Cannot` set `style` of {component_type!r}; got {style!r}.'
        )
    
    
    def iter_components(self):
        """
        Iterates over the direct sub-components of the component.
        
        This method is an iterable generator.
        
        Yields
        ------
        component : ``Component``
        """
        components = self.components
        if (components is not None):
            yield from components
    
    
    def iter_options(self):
        """
        Iterates over the options of the component.
        
        This method is an iterable generator.
        
        Yields
        ------
        option : ``StringSelectOption``
        """
        options = self.options
        if (options is not None):
            yield from options



__all__ = ()

from ...color import Color
from ...field_parsers import (
    nullable_date_time_parser_factory, nullable_entity_parser_factory, nullable_flag_parser_factory,
    nullable_string_parser_factory, preinstanced_parser_factory
)
from ...field_putters import (
    nullable_date_time_optional_putter_factory, nullable_entity_optional_putter_factory,
    nullable_flag_optional_putter_factory, nullable_string_optional_putter_factory, preinstanced_putter_factory,
    url_optional_putter_factory
)
from ...field_validators import (
    nullable_date_time_validator_factory, nullable_entity_validator_factory, nullable_flag_validator_factory,
    preinstanced_validator_factory, url_optional_validator_factory
)

from ..embed_author import EmbedAuthor
from ..embed_field import EmbedField
from ..embed_footer import EmbedFooter
from ..embed_image import EmbedImage
from ..embed_provider import EmbedProvider
from ..embed_thumbnail import EmbedThumbnail
from ..embed_video import EmbedVideo

from .constants import EMBED_DESCRIPTION_LENGTH_MAX, EMBED_TITLE_LENGTH_MAX
from .preinstanced import EmbedType

# author

parse_author = nullable_entity_parser_factory('author', EmbedAuthor)
put_author_into = nullable_entity_optional_putter_factory('author', EmbedAuthor)
validate_author = nullable_entity_validator_factory('author', EmbedAuthor)

# color

parse_color = nullable_flag_parser_factory('color', Color)
put_color_into = nullable_flag_optional_putter_factory('color')
validate_color = nullable_flag_validator_factory('color', Color)

# description

parse_description = nullable_string_parser_factory('description')
put_description_into = nullable_string_optional_putter_factory('description')


def validate_description(description):
    """
    Validates the given embed description.
    
    Parameters
    ----------
    description : `None`, `str`, `object`
        Embed author description.
    
    Returns
    -------
    description : `None`, `str`
    
    Raises
    ------
    TypeError
        - If `description`'s type is incorrect.
    ValueError
        - If `description`'s length is out of the expected range.
    """
    if description is None:
        return None
    
    if not isinstance(description, str):
        description = str(description)
    
    description_length = len(description)
    if description_length == 0:
        return None
    
    if description_length > EMBED_DESCRIPTION_LENGTH_MAX:
        raise ValueError(
            f'`description` length` must be <= {EMBED_DESCRIPTION_LENGTH_MAX}, '
            f'got {description_length}; description = {description!r}.'
        )
    
    return description


# fields

def parse_fields(data):
    """
    Parses the fields out from the given embed data.
    
    Parameters
    ----------
    data : `dict` of (`str`, `object`) items
        User presence data.
    
    Returns
    -------
    fields : `None`, `list` of ``EmbedField``
    """
    field_datas = data.get('fields', None)
    if (field_datas is None) or (not field_datas):
        return None
    
    return [EmbedField.from_data(field_data) for field_data in field_datas]


def put_fields_into(fields, data, defaults):
    """
    Puts the given fields into the given data.
    
    Parameters
    ----------
    fields : `None`, `list` of ``EmbedField``
        Activities.
    data : `dict` of (`str`, `object`) items
        Json serializable dictionary.
    defaults : `bool`
        Whether default values should be included as well.
    
    Returns
    -------
    data : `dict` of (`str`, `object`) items
    """
    if defaults or (fields is not None):
        if fields is None:
            field_datas = []
        else:
            field_datas = [field.to_data(defaults = defaults) for field in fields]
        
        data['fields'] = field_datas
    
    return data


def validate_fields(fields):
    """
    Validates the given fields.
    
    Parameters
    ----------
    fields : `None`, `iterable` of ``EmbedField``
        The fields to validate.
    
    Returns
    -------
    fields : `None`, `list` of ``EmbedField``
    
    Raises
    ------
    TypeError
        - If `fields` type is incorrect.
    """
    if fields is None:
        return None
    
    if getattr(fields, '__iter__', None) is None:
        raise TypeError(
            f'`fields` can be `None`, `iterable` of `{EmbedField.__name__}`, got '
            f'{fields.__class__.__name__}; {fields!r}.'
        )
    
    fields_validated = None
    
    for field in fields:
        if not isinstance(field, EmbedField):
            raise TypeError(
                f'`fields` elements can be `{EmbedField.__name__}`, got '
                f'{field.__class__.__name__}; {field!r}; fields = {fields!r}'
            )
        
        if fields_validated is None:
            fields_validated = []
        
        fields_validated.append(field)
    
    return fields_validated


# footer

parse_footer = nullable_entity_parser_factory('footer', EmbedFooter)
put_footer_into = nullable_entity_optional_putter_factory('footer', EmbedFooter)
validate_footer = nullable_entity_validator_factory('footer', EmbedFooter)

# image

parse_image = nullable_entity_parser_factory('image', EmbedImage)
put_image_into = nullable_entity_optional_putter_factory('image', EmbedImage)
validate_image = nullable_entity_validator_factory('image', EmbedImage)

# provider

parse_provider = nullable_entity_parser_factory('provider', EmbedProvider)
put_provider_into = nullable_entity_optional_putter_factory('provider', EmbedProvider)
validate_provider = nullable_entity_validator_factory('provider', EmbedProvider)

# thumbnail

parse_thumbnail = nullable_entity_parser_factory('thumbnail', EmbedThumbnail)
put_thumbnail_into = nullable_entity_optional_putter_factory('thumbnail', EmbedThumbnail)
validate_thumbnail = nullable_entity_validator_factory('thumbnail', EmbedThumbnail)

# timestamp

parse_timestamp = nullable_date_time_parser_factory('timestamp')
put_timestamp_into = nullable_date_time_optional_putter_factory('timestamp')
validate_timestamp = nullable_date_time_validator_factory('timestamp')

# title

parse_title = nullable_string_parser_factory('title')
put_title_into = nullable_string_optional_putter_factory('title')


def validate_title(title):
    """
    Validates the given embed title.
    
    Parameters
    ----------
    title : `None`, `str`, `object`
        Embed author title.
    
    Returns
    -------
    title : `None`, `str`
    
    Raises
    ------
    TypeError
        - If `title`'s type is incorrect.
    ValueError
        - If `title`'s length is out of the expected range.
    """
    if title is None:
        return None
    
    if not isinstance(title, str):
        title = str(title)
    
    title_length = len(title)
    if title_length == 0:
        return None
    
    if title_length > EMBED_TITLE_LENGTH_MAX:
        raise ValueError(
            f'`title` length` must be <= {EMBED_TITLE_LENGTH_MAX}, '
            f'got {title_length}; title = {title!r}.'
        )
    
    return title

# type

parse_type = preinstanced_parser_factory('type', EmbedType, EmbedType.rich)
put_type_into = preinstanced_putter_factory('type')
validate_type = preinstanced_validator_factory('embed_type', EmbedType)

# url

parse_url = nullable_string_parser_factory('url')
put_url_into = url_optional_putter_factory('url')
validate_url = url_optional_validator_factory('url')

# video

parse_video = nullable_entity_parser_factory('video', EmbedVideo)
put_video_into = nullable_entity_optional_putter_factory('video', EmbedVideo)
validate_video = nullable_entity_validator_factory('video', EmbedVideo)



__all__ = ()

from ...field_parsers import force_string_parser_factory
from ...field_putters import force_string_putter_factory
from ...field_validators import force_string_validator_factory

# distributor

parse_distributor = force_string_parser_factory('distributor')
put_distributor_into = force_string_putter_factory('distributor')
validate_distributor = force_string_validator_factory('distributor', 0, 1024)

# id

parse_id = force_string_parser_factory('id')
put_id_into = force_string_putter_factory('id')
validate_id = force_string_validator_factory('sku_id', 0, 1024)

# sku

parse_sku = force_string_parser_factory('sku')
put_sku_into = force_string_putter_factory('sku')
validate_sku = force_string_validator_factory('sku', 0, 1024)



__all__ = ('GuildJoinRequest',)

import reprlib, warnings

from scarletio import copy_docs

from ...bases import EventBase
from ...core import GUILDS
from ...user import ZEROUSER
from ...utils import DATETIME_FORMAT_CODE

from .fields import (
    parse_actioned_at, parse_actioned_by, parse_created_at, parse_form_responses, parse_guild_id, parse_last_seen_at,
    parse_rejection_reason, parse_status, parse_user, put_actioned_at_into, put_actioned_by_into, put_created_at_into,
    put_form_responses_into, put_guild_id_into, put_last_seen_at_into, put_rejection_reason_into, put_status_into,
    put_user_id_into, put_user_into, validate_actioned_at, validate_actioned_by, validate_created_at,
    validate_form_responses, validate_guild_id, validate_last_seen_at, validate_rejection_reason, validate_status,
    validate_user
)
from .preinstanced import GuildJoinRequestStatus


class GuildJoinRequest(EventBase):
    """
    Guild join request received within both `GUILD_JOIN_REQUEST_CREATE` and `GUILD_JOIN_REQUEST_UPDATE` events.
    
    Attributes
    ----------
    actioned_by : `None`, ``ClientUserBase``
        The user who action the join request.
    actioned_at : `None`, `datetime`
        When the join request was actioned at.
    created_at : `None`, `datetime`
        When the join request was created.
    form_responses : `None`, `tuple` of ``GuildJoinRequestFormResponse``
        The responses the user submitted.
    guild_id : `int`
        The respective guild's identifier.
    last_seen_at : `None`, `datetime`
        When the user was last seen.
    rejection_reason : `None`, `str`
        The reason why the join request was rejected.
    status : ``GuildJoinRequestStatus``
        The join request's status.
    user : ``ClientUserBase``
        The user in context.
    """
    __slots__ = (
        'actioned_by', 'actioned_at', 'created_at', 'form_responses', 'guild_id', 'last_seen_at', 'rejection_reason',
        'status', 'user'
    )
    
    def __new__(
        cls,
        *,
        actioned_by = ...,
        actioned_at = ...,
        created_at = ...,
        form_responses = ...,
        guild_id = ...,
        last_seen_at = ...,
        rejection_reason = ...,
        status = ...,
        user = ...,
    ):
        """
        Creates a new guild join request.
        
        Parameters
        ----------
        actioned_by : `None`, ``ClientUserBase``, Optional (Keyword only)
            The user who action the join request.
        actioned_at : `None`, `datetime`, Optional (Keyword only)
            When the join request was actioned at.
        created_at : None`, `datetime`, Optional (Keyword only)
            When the join request was created.
        form_responses : `None`, `iterable` of ``GuildJoinRequestFormResponse``, Optional (Keyword only)
            The responses the user submitted.
        guild_id : `int`, ``Guild``, Optional (Keyword only)
            The respective guild's identifier.
        last_seen_at : `None`, `datetime`, Optional (Keyword only)
            When the user was last seen.
        rejection_reason : `None`, `str`, Optional (Keyword only)
            The reason why the join request was rejected.
        status : ``GuildJoinRequestStatus``, Optional (Keyword only)
            The join request's status.
        user : ``ClientUserBase``, Optional (Keyword only)
            The user in context.
        
        Raises
        ------
        TypeError
            - If a parameter's type is invalid.
        ValueError
            - If a parameter's value is invalid.
        """
        # actioned_by
        if actioned_by is ...:
            actioned_by = None
        else:
            actioned_by = validate_actioned_by(actioned_by)
        
        # actioned_at
        if actioned_at is ...:
            actioned_at = None
        else:
            actioned_at = validate_actioned_at(actioned_at)
        
        # created_at
        if created_at is ...:
            created_at = None
        else:
            created_at = validate_created_at(created_at)
        
        # form_responses
        if form_responses is ...:
            form_responses = None
        else:
            form_responses = validate_form_responses(form_responses)
        
        # guild_id
        if guild_id is ...:
            guild_id = 0
        else:
            guild_id = validate_guild_id(guild_id)
        
        # last_seen_at
        if last_seen_at is ...:
            last_seen_at = None
        else:
            last_seen_at = validate_last_seen_at(last_seen_at)
        
        # rejection_reason
        if rejection_reason is ...:
            rejection_reason = None
        else:
            rejection_reason = validate_rejection_reason(rejection_reason)
        
        # status
        if status is ...:
            status = GuildJoinRequestStatus.started
        else:
            status = validate_status(status)
        
        # user
        if user is ...:
            user = ZEROUSER
        else:
            user = validate_user(user)
        
        # Construct
        self = object.__new__(cls)
        self.actioned_by = actioned_by
        self.actioned_at = actioned_at
        self.created_at = created_at
        self.form_responses = form_responses
        self.guild_id = guild_id
        self.last_seen_at = last_seen_at
        self.rejection_reason = rejection_reason
        self.status = status
        self.user = user
        return self
    
    
    
    @classmethod
    def from_data(cls, data):
        """
        Creates a new guild join request from the given data.
        
        Parameters
        ----------
        data : `dict` of (`str`, `object`) items
            Received data.
        
        Returns
        -------
        self : `instance<cls>`
        """
        self = object.__new__(cls)
        self.actioned_by = parse_actioned_by(data)
        self.actioned_at = parse_actioned_at(data)
        self.created_at = parse_created_at(data)
        self.form_responses = parse_form_responses(data)
        self.guild_id = parse_guild_id(data)
        self.last_seen_at = parse_last_seen_at(data)
        self.rejection_reason = parse_rejection_reason(data)
        self.status = parse_status(data)
        self.user = parse_user(data)
        return self
    
    
    def to_data(self, *, defaults = False):
        """
        Serializes the guild join request.
        
        Parameters
        ----------
        defaults : `bool` = `False`, Optional (Keyword only)
            Whether fields of their default value should be included as well.
        
        Returns
        -------
        data : `dict` of (`str`, `object`) items
        """
        data = {}
        put_actioned_by_into(self.actioned_by, data, defaults)
        put_actioned_at_into(self.actioned_at, data, defaults)
        put_created_at_into(self.created_at, data, defaults)
        put_form_responses_into(self.form_responses, data, defaults)
        put_guild_id_into(self.guild_id, data, defaults)
        put_last_seen_at_into(self.last_seen_at, data, defaults)
        put_rejection_reason_into(self.rejection_reason, data, defaults)
        put_status_into(self.status, data, defaults)
        put_user_into(self.user, data, defaults)
        put_user_id_into(self.user_id, data, defaults)
        return data
    
    
    @copy_docs(EventBase.__repr__)
    def __repr__(self):
        repr_parts = ['<', self.__class__.__name__]
        
        repr_parts.append(' guild_id = ')
        repr_parts.append(repr(self.guild_id))
        
        repr_parts.append(', user = ')
        repr_parts.append(repr(self.user))
        
        created_at = self.created_at
        if (created_at is not None):
            repr_parts.append(', created_at = ')
            repr_parts.append(format(created_at, DATETIME_FORMAT_CODE))
        
        repr_parts.append(', status = ')
        repr_parts.append(self.status.name)
        
        actioned_by = self.actioned_by
        if (actioned_by is not None):
            repr_parts.append(', actioned_by = ')
            repr_parts.append(repr(actioned_by))
        
        actioned_at = self.actioned_at
        if (actioned_at is not None):
            repr_parts.append(', actioned_at = ')
            repr_parts.append(format(actioned_at, DATETIME_FORMAT_CODE))
        
        last_seen_at = self.last_seen_at
        if (last_seen_at is not None):
            repr_parts.append(', last_seen_at = ')
            repr_parts.append(format(last_seen_at, DATETIME_FORMAT_CODE))
        
        form_responses = self.form_responses
        if (form_responses is not None):
            repr_parts.append(' form_responses = [')
            
            form_response_count = len(form_responses)
            index = 0
            while True:
                form_response = form_responses[index]
                repr_parts.append(repr(form_response))
                
                index += 1
                if index == form_response_count:
                    break
                
                repr_parts.append(', ')
                continue
            
            repr_parts.append(']')
        
        rejection_reason = self.rejection_reason
        if (rejection_reason is not None):
            repr_parts.append(', rejection_reason = ')
            repr_parts.append(reprlib.repr(rejection_reason))
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    @copy_docs(EventBase.__len__)
    def __len__(self):
        return 3
    
    
    @copy_docs(EventBase.__iter__)
    def __iter__(self):
        yield self.guild_id
        yield self.user
        yield self.status
    
    
    @copy_docs(EventBase.__eq__)
    def __eq__(self, other):
        if type(self) is not type(other):
            return NotImplemented
        
        if self.actioned_by is not other.actioned_by:
            return False
        
        if self.actioned_at != other.actioned_at:
            return False
        
        if self.created_at != other.created_at:
            return False
        
        if self.form_responses != other.form_responses:
            return False
        
        if self.guild_id != other.guild_id:
            return False
        
        if self.last_seen_at != other.last_seen_at:
            return False
        
        if self.rejection_reason != other.rejection_reason:
            return False
        
        if self.status != other.status:
            return False
        
        if self.user is not other.user:
            return False
        
        return True
    
    
    @copy_docs(EventBase.__hash__)
    def __hash__(self):
        hash_value = 0
        
        # actioned_by
        actioned_by = self.actioned_by
        if (actioned_by is not None):
            hash_value ^= hash(actioned_by)
        
        # actioned_at
        actioned_at = self.actioned_at
        if (actioned_at is not None):
            hash_value ^= hash(actioned_at)
        
        # created_at
        created_at = self.created_at
        if (created_at is not None):
            hash_value ^= hash(created_at)
        
        # form_responses
        form_responses = self.form_responses
        if (form_responses is not None):
            hash_value ^= len(form_responses)
            
            for form_response in form_responses:
                hash_value ^= hash(form_response)
        
        # guild_id
        hash_value ^= self.guild_id
        
        # last_seen_at
        last_seen_at = self.last_seen_at
        if (last_seen_at is not None):
            hash_value ^= hash(last_seen_at)
        
        # rejection_reason
        rejection_reason = self.rejection_reason
        if (rejection_reason is not None):
            hash_value += hash(rejection_reason)
        
        # status
        hash_value ^= hash(self.status)
        
        # user_id
        hash_value ^= hash(self.user)
        
        return hash_value
    
    
    def copy(self):
        """
        Copies the guild join request.
        
        Returns
        -------
        new : `instance<type<self>>`
        """
        # Construct
        new = object.__new__(type(self))
        new.actioned_by = self.actioned_by
        new.actioned_at = self.actioned_at
        new.created_at = self.created_at
        
        form_responses = self.form_responses
        if (form_responses is not None):
            form_responses = (*(form_response.copy() for form_response in form_responses),)
        new.form_responses = form_responses
        
        new.guild_id = self.guild_id
        new.last_seen_at = self.last_seen_at
        new.rejection_reason = self.rejection_reason
        new.status = self.status
        new.user = self.user
        return new
    
    
    def copy_with(
        self,
        *,
        actioned_by = ...,
        actioned_at = ...,
        created_at = ...,
        form_responses = ...,
        guild_id = ...,
        last_seen_at = ...,
        rejection_reason = ...,
        status = ...,
        user = ...,
    ):
        """
        Copies the guild join request with the given fields.
        
        Parameters
        ----------
        actioned_by : `None`, ``ClientUserBase``, Optional (Keyword only)
            The user who action the join request.
        actioned_at : `None`, `datetime`, Optional (Keyword only)
            When the join request was actioned at.
        created_at : None`, `datetime`, Optional (Keyword only)
            When the join request was created.
        form_responses : `None`, `iterable` of ``GuildJoinRequestFormResponse``, Optional (Keyword only)
            The responses the user submitted.
        guild_id : `int`, ``Guild``, Optional (Keyword only)
            The respective guild's identifier.
        last_seen_at : `None`, `datetime`, Optional (Keyword only)
            When the user was last seen.
        rejection_reason : `None`, `str`, Optional (Keyword only)
            The reason why the join request was rejected.
        status : ``GuildJoinRequestStatus``, Optional (Keyword only)
            The join request's status.
        user : ``ClientUserBase``, Optional (Keyword only)
            The user in context.
        
        Returns
        -------
        new : `instance<type<self>>`
        
        Raises
        ------
        TypeError
            - If a parameter's type is invalid.
        ValueError
            - If a parameter's value is invalid.
        """
        # actioned_by
        if actioned_by is ...:
            actioned_by = self.actioned_by
        else:
            actioned_by = validate_actioned_by(actioned_by)
        
        # actioned_at
        if actioned_at is ...:
            actioned_at = self.actioned_at
        else:
            actioned_at = validate_actioned_at(actioned_at)
        
        # created_at
        if created_at is ...:
            created_at = self.created_at
        else:
            created_at = validate_created_at(created_at)
        
        # form_responses
        if form_responses is ...:
            form_responses = self.form_responses
            if (form_responses is not None):
                form_responses = (*(form_response.copy() for form_response in form_responses),)
        else:
            form_responses = validate_form_responses(form_responses)
        
        # guild_id
        if guild_id is ...:
            guild_id = self.guild_id
        else:
            guild_id = validate_guild_id(guild_id)
        
        # last_seen_at
        if last_seen_at is ...:
            last_seen_at = self.last_seen_at
        else:
            last_seen_at = validate_last_seen_at(last_seen_at)
        
        # rejection_reason
        if rejection_reason is ...:
            rejection_reason = self.rejection_reason
        else:
            rejection_reason = validate_rejection_reason(rejection_reason)
        
        # status
        if status is ...:
            status = self.status
        else:
            status = validate_status(status)
        
        # user
        if user is ...:
            user = self.user
        else:
            user = validate_user(user)
        
        # Construct
        new = object.__new__(type(self))
        new.actioned_by = actioned_by
        new.actioned_at = actioned_at
        new.created_at = created_at
        new.form_responses = form_responses
        new.guild_id = guild_id
        new.last_seen_at = last_seen_at
        new.rejection_reason = rejection_reason
        new.status = status
        new.user = user
        return new
    
    
    @property
    def guild(self):
        """
        Returns the guild join request delete event's guild.
        
        Returns
        -------
        guild : `None`, ``Guild``
        """
        guild_id = self.guild_id
        if guild_id:
            return GUILDS.get(guild_id, None)
    
    
    @property
    def user_id(self):
        """
        Returns the user's identifier who requested to join.
        
        Returns
        -------
        user_id : `int`
        """
        return self.user.id
    
    
    def iter_form_responses(self):
        """
        Iterates over the form responses of the guild join request.
        
        This method is an iterable generator.
        
        Yields
        ------
        form_response : ``GuildJoinRequestFormResponse``
        """
        form_responses = self.form_responses
        if (form_responses is not None):
            yield from form_responses
    
    
    @property
    def last_seen(self):
        """
        ``.last_seen`` is deprecated and will be removed in 2023 April. Please use ``.edited_at`` instead.
        """
        warnings.warn(
            (
                f'`{self.__class__.__name__}.last_seen` is deprecated and will be removed in 2023 November. '
                f'Please use `.last_seen_at` instead.'
            ),
            FutureWarning,
            stacklevel = 2,
        )
        
        return self.last_seen_at



__all__ = ('AutoModerationActionExecutionEvent',)

from scarletio import copy_docs

from ...bases import EventBase
from ...channel import ChannelType, create_partial_channel_from_id
from ...core import AUTO_MODERATION_RULES, GUILDS
from ...message import Message
from ...user import create_partial_user_from_id

from ..action import AutoModerationAction
from ..rule import AutoModerationRuleTriggerType

from .fields import (
    parse_action, parse_alert_system_message_id, parse_channel_id, parse_content, parse_guild_id, parse_matched_content,
    parse_matched_keyword, parse_rule_id, parse_rule_trigger_type, parse_user_id, put_action_into,
    put_alert_system_message_id_into, put_channel_id_into, put_content_into, put_guild_id_into,
    put_matched_content_into, put_matched_keyword_into, put_rule_id_into, put_rule_trigger_type_into, put_user_id_into,
    validate_action, validate_alert_system_message_id, validate_channel_id, validate_content, validate_guild_id,
    validate_matched_content, validate_matched_keyword, validate_rule_id, validate_rule_trigger_type, validate_user_id
)


class AutoModerationActionExecutionEvent(EventBase):
    """
    Represents an auto moderation rule's execution.
    
    Attributes
    ----------
    action : ``AutoModerationAction``
        The action which was executed.
    
    alert_system_message_id : `int`
        The id of the system auto moderation message posted as a result of this action.
        
        > Will default to `0` if the action's type is not `AutoModerationActionType.send_alert_message`.
    
    channel_id : `int`
        The channel's identifier where the user content was posted.
        
        > Defaults to `0` if not applicable.
    
    content : `None`, `str`
        The user generated text content.
    
    guild_id : `int`
        The guild's identifier where the action was executed.
    
    matched_content : `None`, `str`
        The substring in the user submitted content that triggered the rule.
    
    matched_keyword : `None`, `str`
        The matched keyword of the triggered rule.
    
    rule_id : `int`
        The triggered rule's identifier.
    
    rule_trigger_type : ``AutoModerationRuleTriggerType``
        The triggered rule's type.
    
    user_id : `int`
        The user who generated the content triggering the rule.
    """
    __slots__ = (
        'action', 'alert_system_message_id', 'channel_id', 'content', 'guild_id', 'matched_content', 'matched_keyword',
        'rule_id', 'rule_trigger_type', 'user_id'
    )
    
    
    def __new__(
        cls,
        *,
        action = ..., 
        alert_system_message_id = ...,
        channel_id = ...,
        content = ...,
        guild_id = ...,
        matched_content = ...,
        matched_keyword = ...,
        rule_id = ...,
        rule_trigger_type = ...,
        user_id = ...,
    ):
        """
        Creates a new auto mod execution event.
        
        Parameters
        ----------
        action : `None`, ``AutoModerationAction``, Optional (Keyword only)
            The action which was executed.
        
        alert_system_message_id : `None`, ``Message``, `int`, Optional (Keyword only)
            The id of the system auto moderation message posted as a result of this action.
        
        channel_id : `None`, ``Channel``, `int`, Optional (Keyword only)
            The channel's identifier where the user content was posted.
        
        content : `None`, `str`, Optional (Keyword only)
            The user generated text content.
        
        guild_id : `None`, ``Guild``, `int`, Optional (Keyword only)
            The guild's identifier where the action was executed.
        
        matched_content : `None`, `str`, Optional (Keyword only)
            The substring in the user submitted content that triggered the rule.
        
        matched_keyword : `None`, `str`, Optional (Keyword only)
            The matched keyword of the triggered rule.
        
        rule_id : `None`, ``AutoModerationRule``, `int`, Optional (Keyword only)
            The triggered rule's identifier.
        
        rule_trigger_type : `None`, ``AutoModerationRuleTriggerType``, `int`, Optional (Keyword only)
            The triggered rule's type.
        
        user_id : `None`, ``ClientUserBase``, `int`, Optional (Keyword only)
            The user who generated the content triggering the rule.
        
        Returns
        -------
        self : `instance<cls>`
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        # action
        if action is ...:
            action = AutoModerationAction()
        else:
            action = validate_action(action)
        
        # alert_system_message_id
        if alert_system_message_id is ...:
            alert_system_message_id = 0
        else:
            alert_system_message_id = validate_alert_system_message_id(alert_system_message_id)
        
        # channel_id
        if channel_id is ...:
            channel_id = 0
        else:
            channel_id = validate_channel_id(channel_id)
        
        # content
        if content is ...:
            content = None
        else:
            content = validate_content(content)
        
        # guild_id
        if guild_id is ...:
            guild_id = 0
        else:
            guild_id = validate_guild_id(guild_id)
        
        # matched_content
        if matched_content is ...:
            matched_content = None
        else:
            matched_content = validate_matched_content(matched_content)
        
        # matched_keyword
        if matched_keyword is ...:
            matched_keyword = None
        else:
            matched_keyword = validate_matched_keyword(matched_keyword)
        
        # rule_id
        if rule_id is ...:
            rule_id = 0
        else:
            rule_id = validate_rule_id(rule_id)
        
        # rule_trigger_type
        if rule_trigger_type is ...:
            rule_trigger_type = AutoModerationRuleTriggerType.none
        else:
            rule_trigger_type = validate_rule_trigger_type(rule_trigger_type)
        
        # user_id
        if user_id is ...:
            user_id = 0
        else:
            user_id = validate_user_id(user_id)
        
        # Construct
        
        self = object.__new__(cls)
        self.action = action
        self.alert_system_message_id = alert_system_message_id
        self.channel_id = channel_id
        self.content = content
        self.guild_id = guild_id
        self.matched_content = matched_content
        self.matched_keyword = matched_keyword
        self.rule_id = rule_id
        self.rule_trigger_type = rule_trigger_type
        self.user_id = user_id
        return self
    
    
    @classmethod
    def from_data(cls, data):
        """
        Creates a new ``AutoModerationActionExecutionEvent`` from the given data.
        
        Parameters
        ----------
        data : `dict` of (`str`, `Any`) items
            Auto moderation execution event data.
        
        Returns
        -------
        self : `instance<cls>`
        """
        self = object.__new__(cls)
        self.action = parse_action(data)
        self.alert_system_message_id = parse_alert_system_message_id(data)
        self.channel_id = parse_channel_id(data)
        self.content = parse_content(data)
        self.guild_id = parse_guild_id(data)
        self.matched_content = parse_matched_content(data)
        self.matched_keyword = parse_matched_keyword(data)
        self.rule_id = parse_rule_id(data)
        self.rule_trigger_type = parse_rule_trigger_type(data)
        self.user_id = parse_user_id(data)
        return self
    
    
    def to_data(self, *, defaults = False):
        """
        Converts the event to json serializable representation dictionary.
        
        Parameters
        ----------
        defaults : `bool` = `False`, Optional (Keyword only)
            Whether default values should be included as well.
        
        Returns
        -------
        data : `dict` of (`str`, `str`) items
        """
        data = {}
        put_action_into(self.action, data, defaults)
        put_alert_system_message_id_into(self.alert_system_message_id, data, defaults)
        put_channel_id_into(self.channel_id, data, defaults)
        put_content_into(self.content, data, defaults)
        put_guild_id_into(self.guild_id, data, defaults)
        put_matched_content_into(self.matched_content, data, defaults)
        put_matched_keyword_into(self.matched_keyword, data, defaults)
        put_rule_id_into(self.rule_id, data, defaults)
        put_rule_trigger_type_into(self.rule_trigger_type, data, defaults)
        put_user_id_into(self.user_id, data, defaults)
        return data
    
    
    @copy_docs(EventBase.__repr__)
    def __repr__(self):
        repr_parts = ['<', self.__class__.__name__]
        
        # Descriptive fields `.guild_id`, `.channel_id`, `.user_id`, `.rule_id`.
        
        # guild_id
        guild_id = self.guild_id
        if guild_id:
            repr_parts.append(' guild_id = ')
            repr_parts.append(repr(guild_id))
            
            field_added = True
            
        else:
            field_added = False
        
        # channel_id
        channel_id = self.channel_id
        if channel_id:
            if field_added:
                repr_parts.append(',')
            else:
                field_added = True
            
            repr_parts.append(' channel_id = ')
            repr_parts.append(repr(channel_id))
        
        if field_added:
            repr_parts.append(',')
        
        # user_id
        repr_parts.append(' user_id = ')
        repr_parts.append(repr(self.user_id))
        
        # rule_id
        repr_parts.append(', rule_id = ')
        repr_parts.append(repr(self.rule_id))
        
        # Extra fields: `.action`, `.alert_system_message_id`, `.content`, `.matched_content`, `.matched_keyword`
        #    `rule_trigger_type`.
        
        # alert_system_message_id
        alert_system_message_id = self.alert_system_message_id
        if alert_system_message_id:
            repr_parts.append(', alert_system_message_id = ')
            repr_parts.append(repr(alert_system_message_id))
        
        # action
        repr_parts.append(', action = ')
        repr_parts.append(repr(self.action))
        
        # content
        repr_parts.append(', content = ')
        repr_parts.append(repr(self.content))
        
        
        # rule_trigger_type
        rule_trigger_type = self.rule_trigger_type
        repr_parts.append(', rule_trigger_type = ')
        repr_parts.append(repr(rule_trigger_type.name))
        repr_parts.append(' ~ ')
        repr_parts.append(repr(rule_trigger_type.value))
        
        # matched_content
        matched_content = self.matched_content
        if (matched_content is not None):
            repr_parts.append(', matched_content = ')
            repr_parts.append(repr(matched_content))
        
        # matched_keyword
        matched_keyword = self.matched_keyword
        if (matched_keyword is not None):
            repr_parts.append(', matched_keyword = ')
            repr_parts.append(repr(matched_keyword))

        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    @copy_docs(EventBase.__len__)
    def __len__(self):
        return 4
    
    
    @copy_docs(EventBase.__iter__)
    def __iter__(self):
        yield self.guild_id
        yield self.channel_id
        yield self.user_id
        yield self.rule_id
    
    
    @copy_docs(EventBase.__hash__)
    def __hash__(self):
        hash_value = 0
        
        # action
        hash_value ^= hash(self.action)
        
        # alert_system_message_id
        hash_value ^= self.alert_system_message_id
        
        # channel_id
        hash_value ^= self.channel_id
        
        # content
        hash_value ^= hash(self.content)
        
        # guild_id
        hash_value ^= self.guild_id
        
        # matched_content
        matched_content = self.matched_content
        if (matched_content is not None):
            hash_value ^= hash(matched_content)
        
        # matched_keyword
        matched_keyword = self.matched_keyword
        if (matched_keyword is not None):
            hash_value ^= hash(matched_keyword)
        
        # rule_id
        hash_value ^= self.rule_id
        
        # rule_trigger_type
        hash_value ^= self.rule_trigger_type.value
        
        # user_id
        hash_value ^= self.user_id
        
        return hash_value
    
    
    @copy_docs(EventBase.__eq__)
    def __eq__(self, other):
        if type(self) is not type(other):
            return NotImplemented
        
        # action
        if self.action != other.action:
            return False
        
        # alert_system_message_id
        if self.alert_system_message_id != other.alert_system_message_id:
            return False
        
        # channel_id
        if self.channel_id != other.channel_id:
            return False
        
        # content
        if self.content != other.content:
            return False
        
        # guild_id
        if self.guild_id != other.guild_id:
            return False
        
        # matched_content
        if self.matched_content != other.matched_content:
            return False
        
        # matched_keyword
        if self.matched_keyword != other.matched_keyword:
            return False
        
        # rule_id
        if self.rule_id != other.rule_id:
            return False
        
        # rule_trigger_type
        if self.rule_trigger_type != other.rule_trigger_type:
            return False
        
        # user_id
        if self.user_id != other.user_id:
            return False
        
        return True
    
    
    def copy(self):
        """
        Returns a copy of the event.
        
        Returns
        -------
        new : `instance<type<self>>`
        """
        new = object.__new__(type(self))
        new.action = self.action.copy()
        new.alert_system_message_id = self.alert_system_message_id
        new.channel_id = self.channel_id
        new.content = self.content
        new.guild_id = self.guild_id
        new.matched_content = self.matched_content
        new.matched_keyword = self.matched_keyword
        new.rule_id = self.rule_id
        new.rule_trigger_type = self.rule_trigger_type
        new.user_id = self.user_id
        return new
    
    
    def copy_with(
        self,
        *,
        action = ..., 
        alert_system_message_id = ...,
        channel_id = ...,
        content = ...,
        guild_id = ...,
        matched_content = ...,
        matched_keyword = ...,
        rule_id = ...,
        rule_trigger_type = ...,
        user_id = ...,
    ):
        """
        Returns a copy of the event with it's attributes modified based on the defined fields.
        
        Parameters
        ----------
        action : `None`, ``AutoModerationAction``, Optional (Keyword only)
            The action which was executed.
        
        alert_system_message_id : `None`, ``Message``, `int`, Optional (Keyword only)
            The id of the system auto moderation message posted as a result of this action.
        
        channel_id : `None`, ``Channel``, `int`, Optional (Keyword only)
            The channel's identifier where the user content was posted.
        
        content : `None`, `str`, Optional (Keyword only)
            The user generated text content.
        
        guild_id : `None`, ``Guild``, `int`, Optional (Keyword only)
            The guild's identifier where the action was executed.
        
        matched_content : `None`, `str`, Optional (Keyword only)
            The substring in the user submitted content that triggered the rule.
        
        matched_keyword : `None`, `str`, Optional (Keyword only)
            The matched keyword of the triggered rule.
        
        rule_id : `None`, ``AutoModerationRule``, `int`, Optional (Keyword only)
            The triggered rule's identifier.
        
        rule_trigger_type : `None`, ``AutoModerationRuleTriggerType``, `int`, Optional (Keyword only)
            The triggered rule's type.
        
        user_id : `None`, ``ClientUserBase``, `int`, Optional (Keyword only)
            The user who generated the content triggering the rule.
        
        Returns
        -------
        new : `instance<type<self>>`
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        # action
        if action is ...:
            action = self.action.copy()
        else:
            action = validate_action(action)
        
        # alert_system_message_id
        if alert_system_message_id is ...:
            alert_system_message_id = self.alert_system_message_id
        else:
            alert_system_message_id = validate_alert_system_message_id(alert_system_message_id)
        
        # channel_id
        if channel_id is ...:
            channel_id = self.channel_id
        else:
            channel_id = validate_channel_id(channel_id)
        
        # content
        if content is ...:
            content = self.content
        else:
            content = validate_content(content)
        
        # guild_id
        if guild_id is ...:
            guild_id = self.guild_id
        else:
            guild_id = validate_guild_id(guild_id)
        
        # matched_content
        if matched_content is ...:
            matched_content = self.matched_content
        else:
            matched_content = validate_matched_content(matched_content)
        
        # matched_keyword
        if matched_keyword is ...:
            matched_keyword = self.matched_keyword
        else:
            matched_keyword = validate_matched_keyword(matched_keyword)
        
        # rule_id
        if rule_id is ...:
            rule_id = self.rule_id
        else:
            rule_id = validate_rule_id(rule_id)
        
        # rule_trigger_type
        if rule_trigger_type is ...:
            rule_trigger_type = self.rule_trigger_type
        else:
            rule_trigger_type = validate_rule_trigger_type(rule_trigger_type)
        
        # user_id
        if user_id is ...:
            user_id = self.user_id
        else:
            user_id = validate_user_id(user_id)
        
        # Construct
        
        new = object.__new__(type(self))
        new.action = action
        new.alert_system_message_id = alert_system_message_id
        new.channel_id = channel_id
        new.content = content
        new.guild_id = guild_id
        new.matched_content = matched_content
        new.matched_keyword = matched_keyword
        new.rule_id = rule_id
        new.rule_trigger_type = rule_trigger_type
        new.user_id = user_id
        return new
    
    
    @property
    def channel(self):
        """
        Returns the channel where the user content was posted. Returns a partial channel if not cached,
        
        Returns
        -------
        channel : `None`, ``Channel``
        """
        channel_id = self.channel_id
        if channel_id:
            return create_partial_channel_from_id(channel_id, ChannelType.unknown, self.guild_id)
    
    
    @property
    def guild(self):
        """
        Returns the guild where the action was executed. Returns `None` if the guild is not cached.
        
        > The guild must be cached.
        
        Returns
        -------
        guild : `None`, ``Guild``
        """
        guild_id = self.guild_id
        if guild_id:
            return GUILDS.get(guild_id, None)
    
    
    @property
    def user(self):
        """
        Returns the user who generated the content triggering the rule. Returns a partial user if not cached.
        
        Returns
        -------
        user : ``ClientUserBase``
        """
        return create_partial_user_from_id(self.user_id)
    
    
    @property
    def alert_system_message(self):
        """
        Returns the sent alert message. If the message is not cached will return a partial one.
        
        Returns
        -------
        message : `None`, ``Message``
        """
        alert_system_message_id = self.alert_system_message_id
        if alert_system_message_id:
            return Message._create_from_partial_fields(alert_system_message_id, self.channel_id, self.guild_id)
    
    
    @property
    def rule(self):
        """
        Returns the auto moderation rule that executed the event. Returns `None` if the message is not cached.
        
        Returns
        -------
        message : `None`, ``AutoModerationRule``
        """
        rule_id = self.rule_id
        if rule_id:
            return AUTO_MODERATION_RULES.get(rule_id, None)



__all__ = ('ChannelMetadataPrivateBase',)

from scarletio import copy_docs

from .fields import parse_users, put_users_into, validate_users

from .base import ChannelMetadataBase


class ChannelMetadataPrivateBase(ChannelMetadataBase):
    """
    Channel metadata for private channels.
    
    Attributes
    ----------
    users : `list` of ``ClientUserBase``
        The users in the channel.
    
    Class Attributes
    ----------------
    order_group: `int` = `0`
        The channel's order group used when sorting channels.
    """
    __slots__ = ('users',)
    
    
    def __new__(
        cls,
        *,
        users = ...,
    ):
        """
        Creates a new private base channel metadata from the given parameters.
        
        Parameters
        ----------
        users : `iterable` of ``ClientUserBase``, Optional (Keyword only)
            The users in the channel.
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        # users
        if users is ...:
            users = []
        else:
            users = validate_users(users)
        
        # Construct
        self = object.__new__(cls)
        self.users = users
        return self
    
    
    @classmethod
    @copy_docs(ChannelMetadataBase.from_keyword_parameters)
    def from_keyword_parameters(cls, keyword_parameters):
        return cls(
            users = keyword_parameters.pop('users', ...),
        )
    
    
    @copy_docs(ChannelMetadataBase.__hash__)
    def __hash__(self):
        hash_value = ChannelMetadataBase.__hash__(self)
        
        # users
        users = self.users
        hash_value ^= len(users) << 13
        
        for user in users:
            hash_value ^= hash(user)
        
        return hash_value
    
    
    @classmethod
    @copy_docs(ChannelMetadataBase.from_data)
    def from_data(cls, data):
        self = super(ChannelMetadataPrivateBase, cls).from_data(data)
        
        # users
        self.users = parse_users(data)
        
        return self
    
    
    @copy_docs(ChannelMetadataBase._is_equal_same_type)
    def _is_equal_same_type(self, other):
        if self.users != other.users:
            return False
        
        return True
    
    
    @copy_docs(ChannelMetadataBase._get_users)
    def _get_users(self, channel_entity):
        return self.users
    
    
    @copy_docs(ChannelMetadataBase._iter_users)
    def _iter_users(self, channel_entity):
        yield from self.users
    
    
    @copy_docs(ChannelMetadataBase.name)
    def name(self):
        users = self.users
        if len(users) == 2:
            name = f'Direct Message {users[0].full_name} with {users[1].full_name}'
        else:
            name = f'Direct Message (partial)'
        return name
    
    
    @classmethod
    @copy_docs(ChannelMetadataBase._create_empty)
    def _create_empty(cls):
        self = super(ChannelMetadataPrivateBase, cls)._create_empty()
        
        self.users = []
        
        return self
    
    
    @copy_docs(ChannelMetadataBase.copy)
    def copy(self):
        new = object.__new__(type(self))
        new.users = self.users.copy()
        return new
    
    
    def copy_with(
        self,
        *,
        users = ...,
    ):
        """
        Copies the private base channel metadata with the given fields.
        
        Parameters
        ----------
        users : `iterable` of ``ClientUserBase``, Optional (Keyword only)
            The users in the channel.
        
        Returns
        -------
        new : `instance<type<self>>`
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        # users
        if users is ...:
            users = self.users.copy()
        else:
            users = validate_users(users)
        
        # Construct
        self = object.__new__(type(self))
        self.users = users
        return self
    
    
    @copy_docs(ChannelMetadataBase.copy_with_keyword_parameters)
    def copy_with_keyword_parameters(self, keyword_parameters):
        return self.copy_with(
            users = keyword_parameters.pop('users', ...),
        )
    
    
    @copy_docs(ChannelMetadataBase.to_data)
    def to_data(self, *, defaults = False, include_internals = False):
        data = ChannelMetadataBase.to_data(self, defaults = defaults, include_internals = include_internals)
        
        # users
        if include_internals:
            put_users_into(self.users, data, defaults)
        
        return data



__all__ = ('ChannelMetadataGuildDirectory',)

from scarletio import copy_docs

from ...permission import Permission
from ...permission.permission import PERMISSION_MASK_VIEW_CHANNEL, PERMISSION_NONE, PERMISSION_THREAD_AND_VOICE_DENY


from .guild_main_base import ChannelMetadataGuildMainBase


class ChannelMetadataGuildDirectory(ChannelMetadataGuildMainBase):
    """
    Guild directory channel metadata.
    
    Attributes
    ----------
    _cache_permission : `None`, `dict` of (`int`, ``Permission``) items
        A `user_id` to ``Permission`` relation mapping for caching permissions. Defaults to `None`.
    name : `str`
        The channel's name.
    parent_id : `int`
        The channel's parent's identifier.
    permission_overwrites :`None`,  `dict` of (`int`, ``PermissionOverwrite``) items
        The channel's permission overwrites.
    position : `int`
        The channel's position.
    
    Class Attributes
    ----------------
    order_group: `int` = `0`
        The channel's order group used when sorting channels.
    """
    __slots__ = ()
    
    @copy_docs(ChannelMetadataGuildMainBase._get_display_name)
    def _get_display_name(self):
        return self.name.lower()
    
    
    @copy_docs(ChannelMetadataGuildMainBase._get_permissions_for)
    def _get_permissions_for(self, channel_entity, user):
        result = self._get_base_permissions_for(channel_entity, user)
        if not result & PERMISSION_MASK_VIEW_CHANNEL:
            return PERMISSION_NONE
        
        # directory channels do not have thread and voice related permissions
        result &= PERMISSION_THREAD_AND_VOICE_DENY
        
        return Permission(result)
    
    
    @copy_docs(ChannelMetadataGuildMainBase._get_permissions_for_roles)
    def _get_permissions_for_roles(self, channel_entity, roles):
        result = self._get_base_permissions_for_roles(channel_entity, roles)
        if not result & PERMISSION_MASK_VIEW_CHANNEL:
            return PERMISSION_NONE
        
        # directory channels do not have thread and voice related permissions
        result &= PERMISSION_THREAD_AND_VOICE_DENY
        return Permission(result)



__all__ = ()

import sys
from datetime import datetime as DateTime
from shutil import copy2 as copy_file
from os import mkdir as create_directory
from os.path import abspath as absolute_path, dirname as get_parent_directory_path, exists, join as join_paths

from scarletio import Future, LOOP_TIME, Task, TaskGroup, WeakReferer

from ....discord import (
    CLIENTS, DATETIME_FORMAT_CODE, KOKORO, stop_clients, run_console_till_interruption, wait_for_interruption
)
from ...core import register


PROFILING_LIBRARY = None


def _log(message):
    """
    Logs the given message.
    
    Parameters
    ----------
    message : `str`
        The message to log.
    """
    sys.stdout.write(f'{DateTime.utcnow():{DATETIME_FORMAT_CODE}} {message}\n')


class ConnectionEventHandler:
    """
    handler of launch and shutdown event handlers.
    
    Attributes
    ----------
    _client_reference : ``WeakReferer`` to ``Client``
        Reference to the respective client.
    _do_log : `bool`
        Whether actions should be logged.
    _launch_handler : `int`
        Whether the event handler is a launch handler.
    _waiter : ``Future``
        The respective future.
    """
    __slots__ = ('_client_reference', '_do_log', '_launch_handler', '_waiter')
    
    def __new__(cls, client, waiter, launch_handler, do_log):
        """
        Creates a new launch and shutdown event handler.
        
        Parameters
        ----------
        waiter : ``Future``
            The respective future.
        launch_handler : `int`
            Whether the event handler is a launch handler.
        do_log : `bool`
            Whether actions should be logged.
        """
        self = object.__new__(cls)
        self._client_reference = WeakReferer(client)
        self._do_log = do_log
        self._launch_handler = launch_handler
        self._waiter = waiter
        return self
    
    
    def __repr__(self):
        """Returns the event handler's representation."""
        repr_parts = ['<', self.__class__.__name__]
        
        if self.is_cancelled():
            repr_parts.append(' cancelled')
        
        repr_parts.append('shutdown' if self._launch_handler else 'launch')
        repr_parts.append(' event handler')
        
        client = self.get_client()
        if (client is not None):
            repr_parts.append(' of client: ')
            repr_parts.append(repr(client))
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    @property
    def __event_name__(self):
        """
        Returns of which event the event handler is responsible for.
        
        Returns
        -------
        event_name : `str`
        """
        if self._launch_handler:
            event_name = 'launch'
        else:
            event_name = 'shutdown'
        
        return event_name
    
    
    def get_client(self):
        """
        Returns the client to which one is the event handler bound.
        
        Returns
        -------
        client : `None`, ``Client``
        """
        client_reference = self._client_reference
        if (client_reference is not None):
            return client_reference()
    
    
    def cancel(self):
        """
        Cancels the event handler.
        """
        waiter = self._waiter
        if (waiter is None):
            return
        
        self._waiter = None
        
        client = self.get_client()
        if (client is not None):
            client.events.remove(self)
        
        self._client_reference = None
    
    
    def is_cancelled(self):
        """
        Returns whether the event handler is already cancelled.
        """
        return (self._waiter is None)
    
    
    async def __call__(self, client):
        """
        Calls the event handler.
        
        Parameters
        ----------
        client : ``Client``
            The respective client.
        """
        waiter = self._waiter
        if (waiter is None):
            return
            
        waiter.set_result_if_pending(self._launch_handler)
        
        client = self.get_client()
        
        self.cancel()
        
        if (client is not None) and self._do_log:
            _log_client_connection(client, self._launch_handler)


def _log_clients_start_begin(clients):
    """
    Logs when how much clients are being started.
    
    Parameters
    ----------
    clients : `list` of ``Client``
        All the clients to start.
    """
    total_count = len(CLIENTS)
    client_count = len(clients)
    
    if total_count != client_count:
        message = f'Starting {total_count - client_count} / {total_count} clients ({client_count} already running)'
    else:
        message = f'Starting {total_count} bots'
    _log(message)


def _log_clients_start_finish(clients, duration):
    """
    Logs when how much clients started
    
    Parameters
    ----------
    clients : `list` of ``Client``
        All the clients to start.
    duration : `float`
        How long logging in took.
    """
    client_count = len(clients)
    connected_count = sum(client.running for client in clients)
    
    if client_count == connected_count:
        message = f'All clients connected. Took {duration:.02f} seconds.'
    else:
        message = f'{connected_count} / {client_count} clients connected. Took {duration:.02f} seconds.'
    _log(message)


def _log_client_authorization(client, authorized):
    """
    Logs client authorization.
    
    Parameters
    ----------
    client : ``Client``
        The client who's connection was started.
    authorized : `bool`
        Whether the client authorized.
    """
    if authorized:
        message = f'Client authorized: {client.full_name} {client.id}'
    else:
        message = f'Client authorization failed: {client.id}'
    _log(message)


def _log_client_connection(client, launched):
    """
    Logs client connection.
    
    Parameters
    ----------
    client : ``Client``
        The client who's connection was started.
    launched : `int`
        Whether the client launched.
    """
    message = f'Client {"connected" if launched else "connection failed"}: {client.full_name} {client.id}'
    _log(message)


async def _client_start(client, connection_waiter, do_log):
    """
    Starts the given client.
    
    This function is a coroutine.
    
    Parameters
    ----------
    client : ``Client``
        The client to log in.
    connection_waiter : ``Future``
        Future to set result when the connection finished.
    do_log : `bool`
        Whether actions should be logged.
    """
    try:
        try:
            authorized = await client.connect()
        except GeneratorExit:
            raise
    
        if do_log:
            _log_client_authorization(client, authorized)
    except:
        connection_waiter.set_result_if_pending(0)
        raise
    
    else:
        if not authorized:
            connection_waiter.set_result_if_pending(0)


async def _connect_clients(do_log):
    """
    Connects the clients logging their status.
    
    This function is a coroutine.
    
    Parameters
    ----------
    do_log : `bool`
        Whether actions should be logged.
    """
    if do_log:
        start_time = LOOP_TIME()
    
    clients = [client for client in CLIENTS.values() if not client.running]
    
    if do_log:
        _log_clients_start_begin(clients)
    
    connection_waiters = [Future(KOKORO) for client in clients]
    
    event_handlers = []
    
    for client, connection_waiter in zip(clients, connection_waiters):
        Task(KOKORO, _client_start(client, connection_waiter, do_log))
        
        for is_launch_handler in range(2):
            event_handler = ConnectionEventHandler(client, connection_waiter, is_launch_handler, do_log)
            client.events(event_handler)
            event_handlers.append(event_handler)
    
    # clear up reference.
    event_handler = None
    
    try:
        await TaskGroup(KOKORO, connection_waiters).wait_all()
    except:
        stop_clients()
        raise
    
    else:
        connected_count = sum(waiter.get_result() for waiter in connection_waiters)
    
    finally:
        for event_handler in event_handlers:
            event_handler.cancel()
        
        # clear up references.
        connection_waiters = None
        event_handler = None
        event_handlers = None
    
    
    if do_log:
        _log_clients_start_finish(clients, LOOP_TIME() - start_time)
    
    return connected_count


def _check_profiling_available():
    """
    Checks whether profiling library is available.
    
    Returns
    -------
    available : `bool`
    """
    global PROFILING_LIBRARY
    try:
        import yappi
    except ModuleNotFoundError:
        return False
    
    PROFILING_LIBRARY = yappi
    yappi.set_clock_type('cpu')
    return True


def _begin_profiling():
    """
    Begins profiling if available.
    
    Should be called after ``_check_profiling_available``.
    """
    if PROFILING_LIBRARY is None:
        return
    
    PROFILING_LIBRARY.start(builtins = True)


def _stop_profiling():
    """
    Stops profiling if available.
    
    Should be called after ``_begin_profiling``.
    """
    if PROFILING_LIBRARY is None:
        return
    
    directory_path = absolute_path('.profiles')
    
    if not exists(directory_path):
        # If the current directory was deleted dont write logs.
        if not exists(get_parent_directory_path(directory_path)):
            return
        
        create_directory(directory_path)
    
    PROFILING_LIBRARY.stop()
    stats = PROFILING_LIBRARY.get_func_stats()
    stats = PROFILING_LIBRARY.convert2pstats(stats)
    
    file_path_latest = join_paths(directory_path, 'latest.prof')
    file_path_history = join_paths(directory_path, f'{DateTime.utcnow():%Y_%m_%d_%H_%M_%S}.prof')
    
    try:
        stats.dump_stats(file_path_latest)
        copy_file(file_path_latest, file_path_history)
    except FileNotFoundError:
        # No need to panic.
        pass


@register
def run(
    *,
    console: bool = False,
    log: bool = True,
    profile: bool = False,
):
    """
    Starts the created Discord bots.
    
    On keyboard interrupt shuts the bots down. It might take a few seconds.
    
    When `--console` is defined it will start an interactive console after the clients started up.
    When `--log` is defined (default) logging in status will be logged into `sys.stdout`.
    When `--profile`` is defined a `.prof` file will be created on exit under the `.prof` directory.
    To open the latest profile file use `snakeviz .profiles/latest.prof` (example).
    """
    if profile and (not _check_profiling_available()):
        sys.stdout.write(f'Profiling library not available. Please install `yappi`.\n')
        return
    
    try:
        _begin_profiling()
        
        connected_count = KOKORO.run(_connect_clients(log))
        
        if console:
            run_console_till_interruption()
        elif connected_count:
            wait_for_interruption()
    except KeyboardInterrupt as err:
        raise SystemExit from err
    
    finally:
        _stop_profiling()



__all__ = ()

from ...application_command.application_command.constants import (
    APPLICATION_COMMAND_NAME_LENGTH_MAX, APPLICATION_COMMAND_NAME_LENGTH_MIN
)
from ...field_parsers import entity_id_parser_factory, preinstanced_parser_factory
from ...field_putters import entity_id_putter_factory, preinstanced_putter_factory
from ...field_validators import (
    entity_id_validator_factory, entity_validator_factory, force_string_validator_factory,
    nullable_string_array_validator_factory, preinstanced_validator_factory
)
from ...user import ClientUserBase, User, ZEROUSER

# id

parse_id = entity_id_parser_factory('id')
put_id_into = entity_id_putter_factory('id')
validate_id = entity_id_validator_factory('message_interaction_id')


# name & sub_command_name_stack

def parse_name_and_sub_command_name_stack(data):
    """
    Parses out the message interaction's name and sub command name stack from the given data.
    
    Parameters
    ----------
    data : `dict` of (`str`, `object`) items
        Message interaction data.
    
    Returns
    -------
    name : `str`
        Message interaction name.
    sub_command_name_stack : `None`, `tuple` of `str`
        The sub-command-group and sub-command names.
    """
    raw_name = data.get('name', None)
    if raw_name is None:
        name = ''
        sub_command_name_stack = None
    
    else:
        name_split = raw_name.split(' ')
        
        name_split_length = len(name_split)
        if name_split_length == 1:
            name = name_split[0]
            sub_command_name_stack = None
        
        elif name_split_length > 1:
            name = name_split[0]
            sub_command_name_stack = (*name_split[1:],)
        
        else:
            name = ''
            sub_command_name_stack = None
    
    return name, sub_command_name_stack


def put_name_and_sub_command_name_stack_into(name_and_sub_command_name_stack, data, defaults):
    """
    Puts the message interaction's name and sub command name stack into the given `data` json serializable object.
    
    Parameters
    ----------
    name_and_sub_command_name_stack : `tuple` (`str`, `None` | `tuple` of `str`)
       Message interaction name and sub command name stack.
    data : `dict` of (`str`, `object`) items
        Json serializable dictionary.
    defaults : `bool`
        Whether default values should be included as well.
    
    Returns
    -------
    data : `dict` of (`str`, `object`) items
    """
    name, sub_command_name_stack = name_and_sub_command_name_stack
    if sub_command_name_stack is None:
        raw_name = name
    else:
        raw_name = ' '.join([name, *sub_command_name_stack])
    
    data['name'] = raw_name
    return data


validate_name = force_string_validator_factory(
    'name', APPLICATION_COMMAND_NAME_LENGTH_MIN, APPLICATION_COMMAND_NAME_LENGTH_MAX
)

validate_sub_command_name_stack = nullable_string_array_validator_factory('sub_command_name_stack', ordered = False)

# type

parse_type = preinstanced_parser_factory(
    'type', NotImplemented, NotImplemented, include = 'InteractionType', include_default_attribute_name = 'none'
)

put_type_into = preinstanced_putter_factory('type')

validate_type = preinstanced_validator_factory('type', NotImplemented, include = 'InteractionType')

# user

def parse_user(data, guild_id = 0):
    """
    Parses out a user from the given reaction create or delete event data.
    
    Parameters
    ----------
    data : `dict` of (`str`, `object`) items
        Reaction event data.
    guild_id : `int` = `0`, Optional
        The guild's identifier where the event is from.
    
    Returns
    -------
    user : ``ClientUserBase``
    """
    user_data = data.get('user', None)
    if user_data is None:
        return ZEROUSER
    
    return User.from_data(user_data, data.get('member', None), guild_id)


def put_user_into(user, data, defaults, *, guild_id = 0):
    """
    Puts the given user's representation into the given reaction event data.
    
    Parameters
    ----------
    user : ``ClientUserBase``
        The user to serialize.
    data : `dict` of (`str`, `object`) items
        Reaction event data.
    defaults : `bool`
        Whether fields with their default values should be included as well.
    guild_id : `int` = `0`, Optional (Keyword only)
        The guild's identifier where the event is from.
    
    Returns
    -------
    data : `dict` of (`str`, `object`) items
    """
    data['user'] =  user.to_data(defaults = defaults, include_internals = True)
    
    if guild_id:
        try:
            guild_profile = user.guild_profiles[guild_id]
        except KeyError:
            pass
        else:
            data['member'] = guild_profile.to_data(defaults = defaults, include_internals = True)
    
    return data


validate_user = entity_validator_factory('user', ClientUserBase)



__all__ = ()

from scarletio import RichAttributeErrorBaseType

from .fields import (
    parse_application_id, parse_channel_id, parse_guild_id, put_application_id_into, put_channel_id_into,
    put_guild_id_into, validate_application_id, validate_channel_id, validate_guild_id
)


class EmbeddedActivityStateKey(RichAttributeErrorBaseType):
    """
    Used as a key of an embedded activity.
    
    Attributes
    ----------
    application_id : `int`
        The embedded activity's application's identifier.
    channel_id : `int`
        The embedded activity's channel's identifier.
    guild_id : `int`
        The embedded activity's guild's identifier.
    """
    __slots__ = ('application_id', 'channel_id', 'guild_id')
    
    def __new__(cls, guild_id, channel_id, application_id):
        """
        Creates a new embedded activity state.
        
        Parameters
        ----------
        guild_id : `int`, ``Guild``
            The embedded activity's guild or its identifier.
        channel_id : `int`, ``Channel``
            The embedded activity's channel or its identifier.
        application_id : `int`, ``Application``
            The embedded activity's application or its identifier.
        
        Raises
        ------
        TypeError
            - If a parameter's value is incorrect.
        ValueError
            - If a parameter's type is incorrect.
        """
        application_id = validate_application_id(application_id)
        channel_id = validate_channel_id(channel_id)
        guild_id = validate_guild_id(guild_id)
        
        # Construct
        self = object.__new__(cls)
        self.application_id = application_id
        self.channel_id = channel_id
        self.guild_id = guild_id
        return self
    
    
    @classmethod
    def from_data(cls, data, guild_id):
        """
        Creates a new embedded activity key from the given data.
        
        Parameters
        ----------
        data : `dict` of (`str`, `object`) items
            Received data.
        guild_id : `int`
            The guild's identifier where the activity is.
            If given as `0`, will try to pull it from the given data instead.
        
        Returns
        -------
        self : `instance<cls>`
        """
        if not guild_id:
            guild_id = parse_guild_id(data)
        
        channel_id = parse_channel_id(data)
        application_id = parse_application_id(data)
        
        # Construct
        self = object.__new__(cls)
        self.application_id = application_id
        self.channel_id = channel_id
        self.guild_id = guild_id
        return self
    
    
    def to_data(self, *, defaults = False):
        """
        Converts the embedded activity state key into it's json serializable representation.
        
        Parameters
        ----------
        defaults : `bool` = `False`, Optional (Keyword only)
            Whether fields with their default values should be included as well.
        
        Returns
        -------
        data : `dict` of (`str`, `object`) items
        """
        data = {}
        put_application_id_into(self.application_id, data, defaults)
        put_channel_id_into(self.channel_id, data, defaults)
        put_guild_id_into(self.guild_id, data, defaults)
        return data
    
    
    def __repr__(self):
        """Returns the embedded activity key's representation."""
        repr_parts = [self.__class__.__name__, '(']
        
        # guild_id
        repr_parts.append(' guild_id = ')
        repr_parts.append(repr(self.guild_id))
        
        # channel_id
        repr_parts.append(', channel_id = ')
        repr_parts.append(repr(self.channel_id))
        
        # application_id
        repr_parts.append(', application_id = ')
        repr_parts.append(repr(self.application_id))
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    def __hash__(self):
        """Returns the embedded activity key's hash value."""
        hash_value = 0
        
        # application_id
        hash_value ^= self.application_id
        
        # channel_id
        hash_value ^= self.channel_id
        
        # guild_id
        hash_value ^= self.guild_id
        
        return hash_value
    
    
    def __eq__(self, other):
        """Returns whether the two embedded activity keys are equal."""
        if type(self) is not type(other):
            return NotImplemented
        
        # application_id
        if self.application_id != other.application_id:
            return False
        
        # channel_id
        if self.channel_id != other.channel_id:
            return False
        
        # guild_id
        if self.guild_id != other.guild_id:
            return False
        
        return True
    
    
    def __len__(self):
        """Helper for unpacking."""
        return 3
    
    
    def __iter__(self):
        """Unpacks the embedded activity key in order of `guild_id`, `channel_id`, `application_id`."""
        yield self.guild_id
        yield self.channel_id
        yield self.application_id
    
    
    def copy(self):
        """
        Copies the embedded activity key.
        
        Returns
        -------
        new : `instance<type<self>>`
        """
        new = object.__new__(type(self))
        new.application_id = self.application_id
        new.channel_id = self.channel_id
        new.guild_id = self.guild_id
        return new
    
    
    def copy_with(self, *, application_id = ..., channel_id = ..., guild_id = ...):
        """
        Copies the embedded activity key with the given fields.
        
        Parameters
        ----------
        guild_id : `int`, ``Guild``, Optional (Keyword only)
            The embedded activity's guild or its identifier.
        channel_id : `int`, ``Channel``, Optional (Keyword only)
            The embedded activity's channel or its identifier.
        application_id : `int`, ``Application``, Optional (Keyword only)
            The embedded activity's application or its identifier.
        
        Returns
        -------
        new : `instance<type<self>>`
        
        Raises
        ------
        TypeError
            - If a parameter's value is incorrect.
        ValueError
            - If a parameter's type is incorrect.
        """
        # application_id
        if application_id is ...:
            application_id = self.application_id
        else:
            application_id = validate_application_id(application_id)
        
        # channel_id
        if channel_id is ...:
            channel_id = self.channel_id
        else:
            channel_id = validate_channel_id(channel_id)
        
        # guild_id
        if guild_id is ...:
            guild_id = self.guild_id
        else:
            guild_id = validate_guild_id(guild_id)
        
        new = object.__new__(type(self))
        new.application_id = application_id
        new.channel_id = channel_id
        new.guild_id = guild_id
        return new



__all__ = ()

from ...onboarding import OnboardingPrompt

from ..audit_log_change import AuditLogChange

from .shared import convert_nothing, convert_snowflake_array


def convert_prompts(name, data):
    before = data.get('old_value', None)
    if (before is not None):
        if before:
            before = (*(OnboardingPrompt.from_data(data) for data in before),)
        else:
            before = None
    
    after = data.get('new_value', None)
    if (after is not None):
        if after:
            after = (*(OnboardingPrompt.from_data(data) for data in after),)
        else:
            after = None
    
    return AuditLogChange('actions', before, after)


ONBOARDING_SCREEN_CONVERTERS = {
    'default_channel_ids': convert_snowflake_array,
    'enabled': convert_nothing,
    'prompts': convert_prompts,
}



__all__ = ()

from ..activity import Activity, ActivityType
from ..application import Application
from ..events import IntentFlag
from ..field_validators import bool_validator_factory, entity_id_validator_factory, flag_validator_factory
from ..user import ClientUserBase

# activity

def validate_activity(activity):
    """
    Validates the given client activity.
    
    Parameters
    ----------
    activity : ``Activity``
        Client activity.
    
    Returns
    -------
    activity : ``Activity``
    
    Raises
    ------
    TypeError
        - If `activity`'s type is incorrect.
    ValueError
        - if `activity` is a custom activity.
    """
    if not isinstance(activity, Activity):
        raise TypeError(
            f'`activity` can be `{Activity.__name__}`, got {activity.__class__.__name__}; {activity!r}.'
        )
    
    if (activity.type is ActivityType.custom):
        raise ValueError(
            f'`activity` cannot be {ActivityType.custom.name}, got {activity!r}.'
        )
    
    return activity

# additional owners

def validate_additional_owner_ids(additional_owner_ids):
    """
    Validates the given client additional owners.
    
    Parameters
    ----------
    additional_owner_ids : `None`, `int`, ``ClientUserBase``, `iterable` of (`int`, ``ClientUserBase``)
        Additional users to return `True` for by ``Client.is_owner`.
    
    Returns
    -------
    additional_owner_ids : `None`, `set` of `int`.
    
    Raises
    ------
    TypeError
        - If `additional_owner_ids`'s type is incorrect.
    """
    if additional_owner_ids is None:
        return None
    
    if isinstance(additional_owner_ids, ClientUserBase):
        return {additional_owner_ids.id}
    
    if type(additional_owner_ids) is int:
        return {additional_owner_ids}
    
    if isinstance(additional_owner_ids, int):
        return {int(additional_owner_ids)}
    
    if (getattr(additional_owner_ids, '__iter__', None) is None):
        raise TypeError(
            f'`additional_owner_ids` can be `None`, `int`, `{ClientUserBase.__name__}`, `iterable` of'
            f'(`int`, `{ClientUserBase.__name__}`), got {additional_owner_ids.__class__.__name__}; '
            f'{additional_owner_ids!r}.'
        )
    
    additional_owner_ids_validated = None
    
    for additional_owner in additional_owner_ids:
        if type(additional_owner) is int:
            user_id = additional_owner
        
        elif isinstance(additional_owner, int):
            user_id = int(additional_owner)
            
        elif isinstance(additional_owner, ClientUserBase):
            user_id = additional_owner.id
            
        else:
            raise TypeError(
                f'`additional_owner_ids` can contain `int`, `{ClientUserBase.__name__}` elements, got '
                f'{additional_owner.__class__.__name__}; {additional_owner!r}; '
                f'additional_owner_ids = {additional_owner!r}.'
            )
        
        if additional_owner_ids_validated is None:
            additional_owner_ids_validated = set()
            
        additional_owner_ids_validated.add(user_id)
    
    return additional_owner_ids_validated

# application_id

validate_application_id = entity_id_validator_factory('application_id', Application)

# client_id

validate_client_id = entity_id_validator_factory('client_id')

# extensions

def validate_extensions(extensions):
    """
    Validates the given http client debug options.
    
    Parameters
    ----------
    extensions : `None`, `str`, `iterable` of `str`
        Http client debug options.
    
    Returns
    -------
    extensions : `None`, `set` of `str`
    
    Raises
    ------
    TypeError
        - If `extensions`'s type is incorrect.
    """
    if extensions is None:
        return None

    if isinstance(extensions, str):
        if type(extensions) is not str:
            extensions = str(extensions)
        
        return {extensions}
    
    if (getattr(extensions, '__iter__', None) is None):
        raise TypeError(
            f'`extensions` can be `None`, `str`, `iterable` of `str`, got '
            f'{extensions.__class__.__name__}; {extensions!r}.'
        )
    
    extensions_validated = None
    
    for extension in extensions:
        
        if type(extension) is str:
            pass
        elif isinstance(extension, str):
            extension = str(extension)
        else:
            raise TypeError(
                f'{extensions} contains a non `str`, got '
                f'{extensions.__class__.__name__}; {extensions!r}.'
            )
        
        if extensions_validated is None:
            extensions_validated = set()
        
        extensions_validated.add(extension)
    
    return extensions_validated

# http_debug_options

def validate_http_debug_options(http_debug_options):
    """
    Validates the given http client debug options.
    
    Parameters
    ----------
    http_debug_options : `None`, `str`, `iterable` of `str`
        Http client debug options.
    
    Returns
    -------
    http_debug_options : `None`, `set` of `str`
    
    Raises
    ------
    TypeError
        - If `http_debug_options`'s type is incorrect.
    """
    if http_debug_options is None:
        return None

    if isinstance(http_debug_options, str):
        if type(http_debug_options) is not str:
            http_debug_options = str(http_debug_options)
        
        if not http_debug_options.islower():
            http_debug_options = http_debug_options.lower()
        
        return {http_debug_options}
    
    if (getattr(http_debug_options, '__iter__', None) is None):
        raise TypeError(
            f'`http_debug_options` can be `None`, `str`, `iterable` of `str`, got '
            f'{http_debug_options.__class__.__name__}; {http_debug_options!r}.'
        )
    
    http_debug_options_validated = None
    
    for http_debug_option in http_debug_options:
        
        if type(http_debug_option) is str:
            pass
        elif isinstance(http_debug_option, str):
            http_debug_option = str(http_debug_option)
        else:
            raise TypeError(
                f'{http_debug_options} contains a non `str`, got '
                f'{http_debug_options.__class__.__name__}; {http_debug_options!r}.'
            )
        
        if not http_debug_option.islower():
            http_debug_option = http_debug_option.lower()
        
        if http_debug_options_validated is None:
            http_debug_options_validated = set()
        
        http_debug_options_validated.add(http_debug_option)
    
    return http_debug_options_validated

# intents

validate_intents = flag_validator_factory('intents', IntentFlag)

# secret

def validate_secret(secret):
    """
    Validates the given client secret.
    
    Parameters
    ----------
    secret : `str`
        Client secret.
    
    Returns
    -------
    secret : `str`
    
    Raises
    ------
    TypeError
        - If `secret`'s type is incorrect.
    """
    if type(secret) is str:
        pass
    elif isinstance(secret, str):
        secret = str(secret)
    else:
        raise TypeError(
            f'`secret` can be `str`, got `{secret.__class__.__name__}`; {secret!r}.'
        )
    
    return secret

# should_request_users

validate_should_request_users = bool_validator_factory('should_request_users', True)

# shard_count

def validate_shard_count(shard_count):
    """
    Validates the client's shard count.
    
    Parameters
    ----------
    shard_count : `int`
       The client's shard count.
    
    Returns
    -------
    shard_count : `int`
    
    Raises
    ------
    TypeError
        - If `shard_count`'s type is incorrect.
    """
    if (type(shard_count) is int):
        pass
    elif isinstance(shard_count, int):
        shard_count = int(shard_count)
    else:
        raise TypeError(
            f'`shard_count` can be `int`, got {shard_count.__class__.__name__}; {shard_count!r}.'
        )
    
    if shard_count < 0:
        shard_count = 0
    
    # Default shard count to `0` if we received `1`.
    elif shard_count == 1:
        shard_count = 0
    
    return shard_count

# token

def validate_token(token):
    """
    Validates the given client token.
    
    Parameters
    ----------
    token : `str`
        Client token.
    
    Returns
    -------
    token : `str`
    
    Raises
    ------
    TypeError
        - If `token`'s type is incorrect.
    """
    if (type(token) is str):
        pass
    elif isinstance(token, str):
        token = str(token)
    else:
        raise TypeError(
            f'`token` can be `str`, got {token.__class__.__name__}; {token!r}.'
        )
    
    return token



__all__ = ()

APPLICATION_COMMAND_OPTION_NAME_LENGTH_MIN = 1
APPLICATION_COMMAND_OPTION_NAME_LENGTH_MAX = 32
APPLICATION_COMMAND_OPTION_DESCRIPTION_LENGTH_MIN = 2
APPLICATION_COMMAND_OPTION_DESCRIPTION_LENGTH_MAX = 100



__all__ = ()

from importlib import import_module


LAYOUT_DESCRIPTIONS = {
    'package': (
        'Package layout\'s main benefit is that allows easier installation and management of your application.\n'
        'With it you can use package managers like pip or conda simplifying deployment.\n'
        'This also makes it easier to distribute your application between multiple developers.\n'
        '\n'
        'An other benefit is that it helps avoiding naming conflicts (in other words name pollution) when installing\n'
        'since it provides a namespace for your code.\n'
        '\n'
        'Structure:\n'
        '```\n'
        'root\n'
        '├─ .gitignore\n'
        '├─ README.md\n'
        '├─ pyproject.toml\n'
        '└─ { project }\n'
        '    ├─ .env\n'
        '    ├─ __init__.py\n'
        '    ├─ __main__.py\n'
        '    ├─ cli.py\n'
        '    ├─ bots\n'
        '    │   ├─ __init__.py\n'
        '    │   ├─ { bot_names[n] }.py\n'
        '    │   └─ ...\n'
        '    └─ plugins\n'
        '        └─ __init__.py\n'
        '```\n'
    )
}

DEFAULT_LAYOUT = 'package'


MISSING_LAYOUT_APPENDIX = f'Are you sure the name is correct? Perhaps the layout\'s definition is missing?'


def get_project_structure_builder(layout):
    """
    Gets the project structure builder for the given layout.
    
    Parameters
    ----------
    layout : `str`
        The layout to get the builder for.
    
    Returns
    -------
    project_structure_builder : `(str, str, list<str>) -> None`
    
    Raises
    ------
    RuntimeError
        - If layout is missing or if it does not have `create_project_structure` defined.
    """
    try:
        module = import_module(f'{__spec__.parent}.{layout}')
    except ImportError as err:
        raise RuntimeError(
            f'Layout `{layout}` could not be found.\n{MISSING_LAYOUT_APPENDIX}'
        ) from err
    
    try:
        project_structure_builder = module.create_project_structure
    except AttributeError as err:
        raise RuntimeError(
            f'Layout `{layout}` does not have project structure builder defined.\n{MISSING_LAYOUT_APPENDIX}'
        ) from err
    
    return project_structure_builder



__all__ = ('get_plugin_like',)

from re import I as re_ignore_case, compile as re_compile, escape as re_escape


from ..constants import PLUGINS


def get_plugin_like(name):
    """
    Returns the plugin witch matches the given name the most.
    
    Parameters
    ----------
    name : `str`
        An plugin's name.
    
    Returns
    -------
    plugin : ``Plugin``, `None`.
        The matched plugin if any.
    """
    name = name.replace('-', '_').replace(' ', '_')
    
    target_name_length = len(name)
    pattern = re_compile(re_escape(name), re_ignore_case)
    
    accurate_plugin = None
    accurate_name_length = -1
    
    for plugin in PLUGINS.values():
        plugin_name = plugin.name
        name_length = len(plugin_name)
        if (accurate_name_length != -1) and (name_length > accurate_name_length):
            continue
        
        if pattern.search(plugin_name) is None:
            continue
        
        if (accurate_name_length == -1) or (name_length < accurate_name_length):
            accurate_plugin = plugin
            accurate_name_length = name_length
        
        if (name_length == target_name_length) and (name == plugin_name):
            return plugin
        
        continue
    
    return accurate_plugin



from .category import *
from .command import *
from .function import *
from .helpers import *
from .parameter import *
from .parameter_renderer import *
from .parameter_result import *
from .render_constants import *
from .rendering_helpers import *
from .result import *

__all__ = (
    *category.__all__,
    *command.__all__,
    *function.__all__,
    *helpers.__all__,
    *parameter.__all__,
    *parameter_renderer.__all__,
    *parameter_result.__all__,
    *render_constants.__all__,
    *rendering_helpers.__all__,
    *result.__all__,
)



__all__ = ('PluginError', )

from scarletio import CauseGroup

from .constants import (
    PLUGIN_ACTION_FLAG_LOAD, PLUGIN_ACTION_FLAG_NONE, PLUGIN_ACTION_FLAG_SYNTAX_CHECK, PLUGIN_ACTION_FLAG_UNLOAD
)


class PluginError(Exception):
    """
    An exception raised by the ``PluginLoader``, if loading, reloading or unloading an plugin fails with any
    reason.
    
    Attributes
    ----------
    _message : `None`, `str`
        The error's message.
    action : `int`
        Bitwise flags of the actions that were executed.
    plugin : `None`, `list`
        The plugin being actioned when the exception occurred.
    plugin_tree_iterators : `None`, `list` of ``PluginTreeIterator``
        Plugin tree iterators used in the executed actions.
    value : `None`, `object`
        If finding a plugin failed, this value may be set with additional information.
    """
    def __init__(
        self,
        message = None,
        *,
        action = PLUGIN_ACTION_FLAG_NONE,
        cause = None,
        plugin = None,
        plugin_tree_iterators = None,
        value = None,
    ):
        """
        Creates a new plugin error.
        
        Parameters
        ----------
        message : `None`, `str` = `None`, Optional
            The error's message.
        
        action : `int`
            Bitwise flags of the actions that were executed.
        
        cause : `None`, `BaseException` = `None`, Optional (Keyword only)
            Exception cause to apply manually.
        
        plugin : `None`, ``Plugin`` = `None`, Optional (Keyword only)
            The plugin being actioned when the exception occurred.
        
        plugin_tree_iterators : `None`, `list` of ``PluginTreeIterator`` = `None`, Optional (Keyword only)
            Plugin tree iterators used in the executed actions.
        
        value : `None`, `object` = `None`, Optional (Keyword only)
            If finding a plugin failed this value might be set with additional information.
        """
        self._message = message
        self.action = action
        self.plugin = plugin
        self.plugin_tree_iterators = plugin_tree_iterators
        self.value = value
        
        if message is None:
            Exception.__init__(self)
        else:
            Exception.__init__(self, message)
        
        if (cause is not None):
            self.__cause__ = cause
    
    
    def get_plugins(self):
        """
        Returns the plugins that were causing the exception.
        
        Returns
        -------
        plugins : `set` of ``Plugin``
        """
        exceptions_to_scan = [self]
        plugins = set()
        
        while exceptions_to_scan:
            exception = exceptions_to_scan.pop()
            
            if isinstance(exception, PluginError):
                plugin = exception.plugin
                if (plugin is not None):
                    plugins.add(plugin)
            
            cause = exception.__cause__
            if cause is None:
                continue
            
            if isinstance(cause, CauseGroup):
                exceptions_to_scan.extend(cause)
                continue
            
            exceptions_to_scan.append(cause)
            continue
            
        return plugins
    
    
    def get_plugin_tree_iterator_for_syntax_check(self):
        """
        Returns the used plugin tree iterator for syntax check action.
        
        Returns
        -------
        plugin_tree_iterator : `None`, ``PluginTreeIterator``
        """
        return self.get_plugin_tree_iterator_for_action(PLUGIN_ACTION_FLAG_SYNTAX_CHECK)
    
    
    def get_plugin_tree_iterator_for_load(self):
        """
        Returns the used plugin tree iterator for load action.
        
        Returns
        -------
        plugin_tree_iterator : `None`, ``PluginTreeIterator``
        """
        return self.get_plugin_tree_iterator_for_action(PLUGIN_ACTION_FLAG_LOAD)
    
    
    def get_plugin_tree_iterator_for_unload(self):
        """
        Returns the used plugin tree iterator for unload action.
        
        Parameters
        ----------
        action : `int`
            Action flag.
        
        Returns
        -------
        plugin_tree_iterator : `None`, ``PluginTreeIterator``
        """
        return self.get_plugin_tree_iterator_for_action(PLUGIN_ACTION_FLAG_UNLOAD)
    
    
    def get_plugin_tree_iterator_for_action(self, action):
        """
        Returns the used plugin tree iterator for the given action.
        
        Parameters
        ----------
        action : `int`
            Action flag.
        
        Returns
        -------
        plugin_tree_iterator : `None`, ``PluginTreeIterator``
        """
        plugin_tree_iterators = self.plugin_tree_iterators
        if (plugin_tree_iterators is not None):
            for plugin_tree_iterator in plugin_tree_iterators:
                if plugin_tree_iterator.action == action:
                    return plugin_tree_iterator
    
    
    def iter_plugin_tree_iterators(self):
        """
        Iterates over the plugin trees
        """
        plugin_tree_iterators = self.plugin_tree_iterators
        if (plugin_tree_iterators is not None):
            yield from plugin_tree_iterators
    
    
    @property
    def message(self):
        """
        Returns the plugin error's message.
        
        If the plugin error contains more messages connects them.
        
        Returns
        -------
        message : ``Message``
        """
        return '\n\n'.join(self.messages)
    
    
    @property
    def messages(self):
        """
        Returns a list containing the exception error's messages.
        
        Returns
        -------
        messages : `list` of `str`
        """
        messages = []
        
        message = self._message
        if (message is not None):
            messages.append(message)
        
        cause = self.__cause__
        if (cause is not None):
            if isinstance(cause, CauseGroup):
                for cause in cause:
                    messages.append(repr(cause))
            
            else:
                messages.append(repr(cause))
        
        return messages
    
    
    def __len__(self):
        """Returns the amount of messages, what the plugin error contains."""
        length = 0
        
        if (self._message is not None):
            length += 1

        cause = self.__cause__
        if (cause is not None):
            if isinstance(cause, CauseGroup):
                length += len(cause)
            
            else:
                length += 1
        
        return length
    
    
    def __repr__(self):
        """Returns the exception error's representation."""
        repr_parts = ['<', self.__class__.__name__]
        
        message = self._message
        if (message is not None):
            repr_parts.append(' message = ')
            repr_parts.append(repr(message))
        
        cause = self.__cause__
        if (cause is not None):
            if isinstance(cause, CauseGroup):
                cause_count = len(cause)
            
            else:
                cause_count = 1
            
            repr_parts.append(' with ')
            repr_parts.append(repr(cause_count))
            repr_parts.append(' cause')
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    __str__ = __repr__


class DoNotLoadPlugin(BaseException):
    """
    Raised to stop an plugin loaded without error.
    """



__all__ = ('ApplicationCommandOptionType',)

from ...bases import Preinstance as P, PreinstancedBase

from ..application_command_option_metadata import (
    ApplicationCommandOptionMetadataBase, ApplicationCommandOptionMetadataChannel,
    ApplicationCommandOptionMetadataFloat, ApplicationCommandOptionMetadataInteger,
    ApplicationCommandOptionMetadataNested, ApplicationCommandOptionMetadataParameter,
    ApplicationCommandOptionMetadataString, ApplicationCommandOptionMetadataSubCommand
)


class ApplicationCommandOptionType(PreinstancedBase):
    """
    Represents an application command option's type.
    
    Attributes
    ----------
    name : `str`
        The name of the application command option type.
    value : `int`
        The identifier value the application command option type.
    metadata_type : `type<ApplicationCommandOptionMetadataBase>`
        The option type's respective metadata type.
    
    Class Attributes
    ----------------
    INSTANCES : `dict` of (`int`, ``ApplicationCommandOptionType``) items
        Stores the predefined ``ApplicationCommandOptionType``-s. These can be accessed with their `value` as
        key.
    VALUE_TYPE : `type` = `int`
        The application command option types' values' type.
    DEFAULT_NAME : `str` = `'UNDEFINED'`
        The default name of the application command option types.
    
    Every predefined application command option type can be accessed as class attribute as well:
    
    +-----------------------+-------------------+-------+---------------------------------------------------+
    | Class attribute name  | Name              | Value | Metadata type                                     |
    +=======================+===================+=======+===================================================+
    | none                  | none              | 0     | ``ApplicationCommandOptionMetadataBase``          |
    +-----------------------+-------------------+-------+---------------------------------------------------+
    | sub_command           | sub-command       | 1     | ``ApplicationCommandOptionMetadataSubCommand``    |
    +-----------------------+-------------------+-------+---------------------------------------------------+
    | sub_command_group     | sub-command group | 2     | ``ApplicationCommandOptionMetadataNested``        |
    +-----------------------+-------------------+-------+---------------------------------------------------+
    | string                | string            | 3     | ``ApplicationCommandOptionMetadataString``        |
    +-----------------------+-------------------+-------+---------------------------------------------------+
    | integer               | integer           | 4     | ``ApplicationCommandOptionMetadataInteger``       |
    +-----------------------+-------------------+-------+---------------------------------------------------+
    | boolean               | boolean           | 5     | ``ApplicationCommandOptionMetadataParameter``     |
    +-----------------------+-------------------+-------+---------------------------------------------------+
    | user                  | user              | 6     | ``ApplicationCommandOptionMetadataParameter``     |
    +-----------------------+-------------------+-------+---------------------------------------------------+
    | channel               | channel           | 7     | ``ApplicationCommandOptionMetadataChannel``       |
    +-----------------------+-------------------+-------+---------------------------------------------------+
    | role                  | role              | 8     | ``ApplicationCommandOptionMetadataParameter``     |
    +-----------------------+-------------------+-------+---------------------------------------------------+
    | mentionable           | mentionable       | 9     | ``ApplicationCommandOptionMetadataParameter``     |
    +-----------------------+-------------------+-------+---------------------------------------------------+
    | float                 | float             | 10    | ``ApplicationCommandOptionMetadataFloat``         |
    +-----------------------+-------------------+-------+---------------------------------------------------+
    | attachment            | attachment        | 11    | ``ApplicationCommandOptionMetadataParameter``     |
    +-----------------------+-------------------+-------+---------------------------------------------------+
    """
    INSTANCES = {}
    VALUE_TYPE = int
    DEFAULT_NAME = 'UNDEFINED'
    
    __slots__ = ('metadata_type',)
    

    @classmethod
    def _from_value(cls, value):
        """
        Creates a new application command option type with the given value.
        
        Parameters
        ----------
        value : `int`
            The application command option type's identifier value.
        
        Returns
        -------
        self : `instance<cls>`
            The created instance.
        """
        self = object.__new__(cls)
        self.name = cls.DEFAULT_NAME
        self.value = value
        self.metadata_type = ApplicationCommandOptionMetadataBase
        
        return self
    
    
    def __init__(self, value, name, metadata_type):
        """
        Creates a new application command option type and stores it at the class's `.INSTANCES` class attribute as well.
        
        Parameters
        ----------
        value : `int`
            The Discord side identifier value of the application command option type.
        name : `str`
            The default name of the application command option type.
        metadata_type : `type<ApplicationCommandOptionMetadataBase>`
            The option type's respective metadata type.
        """
        self.value = value
        self.name = name
        self.metadata_type = metadata_type
        
        self.INSTANCES[value] = self
    
    
    none = P(0, 'none', ApplicationCommandOptionMetadataBase)
    sub_command = P(1, 'sub-command', ApplicationCommandOptionMetadataSubCommand)
    sub_command_group = P(2, 'sub-command group', ApplicationCommandOptionMetadataNested)
    string = P(3, 'string', ApplicationCommandOptionMetadataString)
    integer = P(4, 'integer', ApplicationCommandOptionMetadataInteger)
    boolean = P(5, 'boolean', ApplicationCommandOptionMetadataParameter)
    user = P(6, 'user', ApplicationCommandOptionMetadataParameter)
    channel = P(7, 'channel', ApplicationCommandOptionMetadataChannel)
    role = P(8, 'role', ApplicationCommandOptionMetadataParameter)
    mentionable = P(9, 'mentionable', ApplicationCommandOptionMetadataParameter)
    float = P(10, 'float', ApplicationCommandOptionMetadataFloat)
    attachment = P(11, 'attachment', ApplicationCommandOptionMetadataParameter)
